import enum

from pyarrow.lib import Buffer

class DlpackDeviceType(enum.IntEnum):
    CPU = 1
    CUDA = 2
    CPU_PINNED = 3
    OPENCL = 4
    VULKAN = 7
    METAL = 8
    VPI = 9
    ROCM = 10

class _PyArrowBuffer:
    def __init__(self, x: Buffer, allow_copy: bool = True) -> None: ...
    @property
    def bufsize(self) -> int: ...
    @property
    def ptr(self) -> int: ...
    def __dlpack__(self): ...
    def __dlpack_device__(self) -> tuple[DlpackDeviceType, int | None]: ...



import enum

from typing import Any, Iterable, TypeAlias, TypedDict

from pyarrow.lib import Array, ChunkedArray

from .buffer import _PyArrowBuffer

class DtypeKind(enum.IntEnum):
    INT = 0
    UINT = 1
    FLOAT = 2
    BOOL = 20
    STRING = 21  # UTF-8
    DATETIME = 22
    CATEGORICAL = 23

Dtype: TypeAlias = tuple[DtypeKind, int, str, str]

class ColumnNullType(enum.IntEnum):
    NON_NULLABLE = 0
    USE_NAN = 1
    USE_SENTINEL = 2
    USE_BITMASK = 3
    USE_BYTEMASK = 4

class ColumnBuffers(TypedDict):
    data: tuple[_PyArrowBuffer, Dtype]
    validity: tuple[_PyArrowBuffer, Dtype] | None
    offsets: tuple[_PyArrowBuffer, Dtype] | None

class CategoricalDescription(TypedDict):
    is_ordered: bool
    is_dictionary: bool
    categories: _PyArrowColumn | None

class Endianness(enum.Enum):
    LITTLE = "<"
    BIG = ">"
    NATIVE = "="
    NA = "|"

class NoBufferPresent(Exception): ...

class _PyArrowColumn:
    def __init__(self, column: Array | ChunkedArray, allow_copy: bool = True) -> None: ...
    def size(self) -> int: ...
    @property
    def offset(self) -> int: ...
    @property
    def dtype(self) -> tuple[DtypeKind, int, str, str]: ...
    @property
    def describe_categorical(self) -> CategoricalDescription: ...
    @property
    def describe_null(self) -> tuple[ColumnNullType, Any]: ...
    @property
    def null_count(self) -> int: ...
    @property
    def metadata(self) -> dict[str, Any]: ...
    def num_chunks(self) -> int: ...
    def get_chunks(self, n_chunks: int | None = None) -> Iterable[_PyArrowColumn]: ...
    def get_buffers(self) -> ColumnBuffers: ...



from typing import IO, Any, Iterable, Iterator, Literal, Sequence, TypeAlias, TypedDict

from _typeshed import StrPath

from ._stubs_typing import Order
from .lib import (
    Buffer,
    ChunkedArray,
    KeyValueMetadata,
    MemoryPool,
    NativeFile,
    RecordBatch,
    Schema,
    Table,
    _Weakrefable,
)

_PhysicalType: TypeAlias = Literal[
    "BOOLEAN",
    "INT32",
    "INT64",
    "INT96",
    "FLOAT",
    "DOUBLE",
    "BYTE_ARRAY",
    "FIXED_LEN_BYTE_ARRAY",
    "UNKNOWN",
]
_LogicTypeName: TypeAlias = Literal[
    "UNDEFINED",
    "STRING",
    "MAP",
    "LIST",
    "ENUM",
    "DECIMAL",
    "DATE",
    "TIME",
    "TIMESTAMP",
    "INT",
    "FLOAT16",
    "JSON",
    "BSON",
    "UUID",
    "NONE",
    "UNKNOWN",
]
_ConvertedType: TypeAlias = Literal[
    "NONE",
    "UTF8",
    "MAP",
    "MAP_KEY_VALUE",
    "LIST",
    "ENUM",
    "DECIMAL",
    "DATE",
    "TIME_MILLIS",
    "TIME_MICROS",
    "TIMESTAMP_MILLIS",
    "TIMESTAMP_MICROS",
    "UINT_8",
    "UINT_16",
    "UINT_32",
    "UINT_64",
    "INT_8",
    "INT_16",
    "INT_32",
    "INT_64",
    "JSON",
    "BSON",
    "INTERVAL",
    "UNKNOWN",
]
_Encoding: TypeAlias = Literal[
    "PLAIN",
    "PLAIN_DICTIONARY",
    "RLE",
    "BIT_PACKED",
    "DELTA_BINARY_PACKED",
    "DELTA_LENGTH_BYTE_ARRAY",
    "DELTA_BYTE_ARRAY",
    "RLE_DICTIONARY",
    "BYTE_STREAM_SPLIT",
    "UNKNOWN",
]
_Compression: TypeAlias = Literal[
    "UNCOMPRESSED",
    "SNAPPY",
    "GZIP",
    "LZO",
    "BROTLI",
    "LZ4",
    "ZSTD",
    "UNKNOWN",
]

class _Statistics(TypedDict):
    has_min_max: bool
    min: Any | None
    max: Any | None
    null_count: int | None
    distinct_count: int | None
    num_values: int
    physical_type: _PhysicalType

class Statistics(_Weakrefable):
    def to_dict(self) -> _Statistics: ...
    def equals(self, other: Statistics) -> bool: ...
    @property
    def has_min_max(self) -> bool: ...
    @property
    def hash_null_count(self) -> bool: ...
    @property
    def has_distinct_count(self) -> bool: ...
    @property
    def min_raw(self) -> Any | None: ...
    @property
    def max_raw(self) -> Any | None: ...
    @property
    def min(self) -> Any | None: ...
    @property
    def max(self) -> Any | None: ...
    @property
    def null_count(self) -> int | None: ...
    @property
    def distinct_count(self) -> int | None: ...
    @property
    def num_values(self) -> int: ...
    @property
    def physical_type(self) -> _PhysicalType: ...
    @property
    def logical_type(self) -> ParquetLogicalType: ...
    @property
    def converted_type(self) -> _ConvertedType | None: ...

class ParquetLogicalType(_Weakrefable):
    def to_json(self) -> str: ...
    @property
    def type(self) -> _LogicTypeName: ...

class _ColumnChunkMetaData(TypedDict):
    file_offset: int
    file_path: str | None
    physical_type: _PhysicalType
    num_values: int
    path_in_schema: str
    is_stats_set: bool
    statistics: Statistics | None
    compression: _Compression
    encodings: tuple[_Encoding, ...]
    has_dictionary_page: bool
    dictionary_page_offset: int | None
    data_page_offset: int
    total_compressed_size: int
    total_uncompressed_size: int

class ColumnChunkMetaData(_Weakrefable):
    def to_dict(self) -> _ColumnChunkMetaData: ...
    def equals(self, other: ColumnChunkMetaData) -> bool: ...
    @property
    def file_offset(self) -> int: ...
    @property
    def file_path(self) -> str | None: ...
    @property
    def physical_type(self) -> _PhysicalType: ...
    @property
    def num_values(self) -> int: ...
    @property
    def path_in_schema(self) -> str: ...
    @property
    def is_stats_set(self) -> bool: ...
    @property
    def statistics(self) -> Statistics | None: ...
    @property
    def compression(self) -> _Compression: ...
    @property
    def encodings(self) -> tuple[_Encoding, ...]: ...
    @property
    def has_dictionary_page(self) -> bool: ...
    @property
    def dictionary_page_offset(self) -> int | None: ...
    @property
    def data_page_offset(self) -> int: ...
    @property
    def has_index_page(self) -> bool: ...
    @property
    def index_page_offset(self) -> int: ...
    @property
    def total_compressed_size(self) -> int: ...
    @property
    def total_uncompressed_size(self) -> int: ...
    @property
    def has_offset_index(self) -> bool: ...
    @property
    def has_column_index(self) -> bool: ...

class _SortingColumn(TypedDict):
    column_index: int
    descending: bool
    nulls_first: bool

class SortingColumn:
    def __init__(
        self, column_index: int, descending: bool = False, nulls_first: bool = False
    ) -> None: ...
    @classmethod
    def from_ordering(
        cls,
        schema: Schema,
        sort_keys: Sequence[tuple[str, Order]],
        null_placement: Literal["at_start", "at_end"] = "at_end",
    ) -> tuple[SortingColumn, ...]: ...
    @staticmethod
    def to_ordering(
        schema: Schema, sorting_columns: tuple[SortingColumn, ...]
    ) -> tuple[Sequence[tuple[str, Order]], Literal["at_start", "at_end"]]: ...
    def __hash__(self) -> int: ...
    @property
    def column_index(self) -> int: ...
    @property
    def descending(self) -> bool: ...
    @property
    def nulls_first(self) -> bool: ...
    def to_dict(self) -> _SortingColumn: ...

class _RowGroupMetaData(TypedDict):
    num_columns: int
    num_rows: int
    total_byte_size: int
    columns: list[ColumnChunkMetaData]
    sorting_columns: list[SortingColumn]

class RowGroupMetaData(_Weakrefable):
    def __init__(self, parent: FileMetaData, index: int) -> None: ...
    def equals(self, other: RowGroupMetaData) -> bool: ...
    def column(self, i: int) -> ColumnChunkMetaData: ...
    def to_dict(self) -> _RowGroupMetaData: ...
    @property
    def num_columns(self) -> int: ...
    @property
    def num_rows(self) -> int: ...
    @property
    def total_byte_size(self) -> int: ...
    @property
    def sorting_columns(self) -> list[SortingColumn]: ...

class _FileMetaData(TypedDict):
    created_by: str
    num_columns: int
    num_rows: int
    num_row_groups: int
    format_version: str
    serialized_size: int

class FileMetaData(_Weakrefable):
    def __hash__(self) -> int: ...
    def to_dict(self) -> _FileMetaData: ...
    def equals(self, other: FileMetaData) -> bool: ...
    @property
    def schema(self) -> ParquetSchema: ...
    @property
    def serialized_size(self) -> int: ...
    @property
    def num_columns(self) -> int: ...
    @property
    def num_rows(self) -> int: ...
    @property
    def num_row_groups(self) -> int: ...
    @property
    def format_version(self) -> str: ...
    @property
    def created_by(self) -> str: ...
    @property
    def metadata(self) -> dict[bytes, bytes] | None: ...
    def row_group(self, i: int) -> RowGroupMetaData: ...
    def set_file_path(self, path: str) -> None: ...
    def append_row_groups(self, other: FileMetaData) -> None: ...
    def write_metadata_file(self, where: StrPath | Buffer | NativeFile | IO) -> None: ...

class ParquetSchema(_Weakrefable):
    def __init__(self, container: FileMetaData) -> None: ...
    def __getitem__(self, i: int) -> ColumnChunkMetaData: ...
    def __hash__(self) -> int: ...
    def __len__(self) -> int: ...
    @property
    def names(self) -> list[str]: ...
    def to_arrow_schema(self) -> Schema: ...
    def equals(self, other: ParquetSchema) -> bool: ...
    def column(self, i: int) -> ColumnSchema: ...

class ColumnSchema(_Weakrefable):
    def __init__(self, schema: ParquetSchema, index: int) -> None: ...
    def equals(self, other: ColumnSchema) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def path(self) -> str: ...
    @property
    def max_definition_level(self) -> int: ...
    @property
    def max_repetition_level(self) -> int: ...
    @property
    def physical_type(self) -> _PhysicalType: ...
    @property
    def logical_type(self) -> ParquetLogicalType: ...
    @property
    def converted_type(self) -> _ConvertedType | None: ...
    @property
    def length(self) -> int | None: ...
    @property
    def precision(self) -> int | None: ...
    @property
    def scale(self) -> int | None: ...

class ParquetReader(_Weakrefable):
    def __init__(self, memory_pool: MemoryPool | None = None) -> None: ...
    def open(
        self,
        source: StrPath | NativeFile | IO,
        *,
        use_memory_map: bool = False,
        read_dictionary: Iterable[int] | Iterable[str] | None = None,
        metadata: FileMetaData | None = None,
        buffer_size: int = 0,
        pre_buffer: bool = False,
        coerce_int96_timestamp_unit: str | None = None,
        decryption_properties: FileDecryptionProperties | None = None,
        thrift_string_size_limit: int | None = None,
        thrift_container_size_limit: int | None = None,
        page_checksum_verification: bool = False,
    ): ...
    @property
    def column_paths(self) -> list[str]: ...
    @property
    def metadata(self) -> FileMetaData: ...
    @property
    def schema_arrow(self) -> Schema: ...
    @property
    def num_row_groups(self) -> int: ...
    def set_use_threads(self, use_threads: bool) -> None: ...
    def set_batch_size(self, batch_size: int) -> None: ...
    def iter_batches(
        self,
        batch_size: int,
        row_groups: list[int],
        column_indices: list[int] | None = None,
        use_threads: bool = True,
    ) -> Iterator[RecordBatch]: ...
    def read_row_group(
        self, i: int, column_indices: list[int] | None = None, use_threads: bool = True
    ) -> Table: ...
    def read_row_groups(
        self,
        row_groups: list[int],
        column_indices: list[int] | None = None,
        use_threads: bool = True,
    ) -> Table: ...
    def read_all(
        self, column_indices: list[int] | None = None, use_threads: bool = True
    ) -> Table: ...
    def scan_contents(self, column_indices: list[int] | None = None, batch_size: int = 65536): ...
    def column_name_idx(self, column_name: str) -> int: ...
    def read_column(self, column_index: int) -> ChunkedArray: ...
    def close(self) -> None: ...
    @property
    def closed(self) -> bool: ...

class ParquetWriter(_Weakrefable):
    def __init__(
        self,
        where: StrPath | NativeFile | IO,
        schema: Schema,
        use_dictionary: bool | list[str] | None = None,
        compression: _Compression | dict[str, _Compression] | None = None,
        version: str | None = None,
        write_statistics: bool | list[str] | None = None,
        memory_pool: MemoryPool | None = None,
        use_deprecated_int96_timestamps: bool = False,
        coerce_timestamps: Literal["ms", "us"] | None = None,
        data_page_size: int | None = None,
        allow_truncated_timestamps: bool = False,
        compression_level: int | dict[str, int] | None = None,
        use_byte_stream_split: bool | list[str] = False,
        column_encoding: _Encoding | dict[str, _Encoding] | None = None,
        writer_engine_version: str | None = None,
        data_page_version: str | None = None,
        use_compliant_nested_type: bool = True,
        encryption_properties: FileDecryptionProperties | None = None,
        write_batch_size: int | None = None,
        dictionary_pagesize_limit: int | None = None,
        store_schema: bool = True,
        write_page_index: bool = False,
        write_page_checksum: bool = False,
        sorting_columns: tuple[SortingColumn, ...] | None = None,
        store_decimal_as_integer: bool = False,
    ): ...
    def close(self) -> None: ...
    def write_table(self, table: Table, row_group_size: int | None = None) -> None: ...
    def add_key_value_metadata(self, key_value_metadata: KeyValueMetadata) -> None: ...
    @property
    def metadata(self) -> FileMetaData: ...
    @property
    def use_dictionary(self) -> bool | list[str] | None: ...
    @property
    def use_deprecated_int96_timestamps(self) -> bool: ...
    @property
    def use_byte_stream_split(self) -> bool | list[str]: ...
    @property
    def column_encoding(self) -> _Encoding | dict[str, _Encoding] | None: ...
    @property
    def coerce_timestamps(self) -> Literal["ms", "us"] | None: ...
    @property
    def allow_truncated_timestamps(self) -> bool: ...
    @property
    def compression(self) -> _Compression | dict[str, _Compression] | None: ...
    @property
    def compression_level(self) -> int | dict[str, int] | None: ...
    @property
    def data_page_version(self) -> str | None: ...
    @property
    def use_compliant_nested_type(self) -> bool: ...
    @property
    def version(self) -> str | None: ...
    @property
    def write_statistics(self) -> bool | list[str] | None: ...
    @property
    def writer_engine_version(self) -> str: ...
    @property
    def row_group_size(self) -> int: ...
    @property
    def data_page_size(self) -> int: ...
    @property
    def encryption_properties(self) -> FileDecryptionProperties: ...
    @property
    def write_batch_size(self) -> int: ...
    @property
    def dictionary_pagesize_limit(self) -> int: ...
    @property
    def store_schema(self) -> bool: ...
    @property
    def store_decimal_as_integer(self) -> bool: ...

class FileEncryptionProperties: ...
class FileDecryptionProperties: ...



from pyarrow._csv import (
    ISO8601,
    ConvertOptions,
    CSVStreamingReader,
    CSVWriter,
    InvalidRow,
    ParseOptions,
    ReadOptions,
    WriteOptions,
    open_csv,
    read_csv,
    write_csv,
)

__all__ = [
    "ISO8601",
    "ConvertOptions",
    "CSVStreamingReader",
    "CSVWriter",
    "InvalidRow",
    "ParseOptions",
    "ReadOptions",
    "WriteOptions",
    "open_csv",
    "read_csv",
    "write_csv",
]



def encode_file_path(path: str | bytes) -> bytes: ...
def tobytes(o: str | bytes) -> bytes: ...
def frombytes(o: bytes, *, safe: bool = False): ...

__all__ = ["encode_file_path", "tobytes", "frombytes"]



from typing import Any

import cuda  # type: ignore[import-not-found]

from numba.cuda.cudadrv import driver as _numba_driver  # type: ignore[import-not-found]

from . import lib
from ._stubs_typing import ArrayLike

class Context(lib._Weakrefable):
    def __init__(self, device_number: int = 0, handle: int | None = None) -> None: ...
    @staticmethod
    def from_numba(context: _numba_driver.Context | None = None) -> Context: ...
    def to_numba(self) -> _numba_driver.Context: ...
    @staticmethod
    def get_num_devices() -> int: ...
    @property
    def device_number(self) -> int: ...
    @property
    def handle(self) -> int: ...
    def synchronize(self) -> None: ...
    @property
    def bytes_allocated(self) -> int: ...
    def get_device_address(self, address: int) -> int: ...
    def new_buffer(self, nbytes: int) -> CudaBuffer: ...
    def foreign_buffer(self, address: int, size: int, base: Any | None = None) -> CudaBuffer: ...
    def open_ipc_buffer(self, ipc_handle: IpcMemHandle) -> CudaBuffer: ...
    def buffer_from_data(
        self,
        data: CudaBuffer | HostBuffer | lib.Buffer | ArrayLike,
        offset: int = 0,
        size: int = -1,
    ) -> CudaBuffer: ...
    def buffer_from_object(self, obj: Any) -> CudaBuffer: ...

class IpcMemHandle(lib._Weakrefable):
    @staticmethod
    def from_buffer(opaque_handle: lib.Buffer) -> IpcMemHandle: ...
    def serialize(self, pool: lib.MemoryPool | None = None) -> lib.Buffer: ...

class CudaBuffer(lib.Buffer):
    @staticmethod
    def from_buffer(buf: lib.Buffer) -> CudaBuffer: ...
    @staticmethod
    def from_numba(mem: _numba_driver.MemoryPointer) -> CudaBuffer: ...
    def to_numba(self) -> _numba_driver.MemoryPointer: ...
    def copy_to_host(
        self,
        position: int = 0,
        nbytes: int = -1,
        buf: lib.Buffer | None = None,
        memory_pool: lib.MemoryPool | None = None,
        resizable: bool = False,
    ) -> lib.Buffer: ...
    def copy_from_host(
        self, data: lib.Buffer | ArrayLike, position: int = 0, nbytes: int = -1
    ) -> int: ...
    def copy_from_device(self, buf: CudaBuffer, position: int = 0, nbytes: int = -1) -> int: ...
    def export_for_ipc(self) -> IpcMemHandle: ...
    @property
    def context(self) -> Context: ...
    def slice(self, offset: int = 0, length: int | None = None) -> CudaBuffer: ...
    def to_pybytes(self) -> bytes: ...

class HostBuffer(lib.Buffer):
    @property
    def size(self) -> int: ...

class BufferReader(lib.NativeFile):
    def __init__(self, obj: CudaBuffer) -> None: ...
    def read_buffer(self, nbytes: int | None = None) -> CudaBuffer: ...

class BufferWriter(lib.NativeFile):
    def __init__(self, obj: CudaBuffer) -> None: ...
    def writeat(self, position: int, data: ArrayLike) -> None: ...
    @property
    def buffer_size(self) -> int: ...
    @buffer_size.setter
    def buffer_size(self, buffer_size: int): ...
    @property
    def num_bytes_buffered(self) -> int: ...

def new_host_buffer(size: int, device: int = 0) -> HostBuffer: ...
def serialize_record_batch(batch: lib.RecordBatch, ctx: Context) -> CudaBuffer: ...
def read_message(
    source: CudaBuffer | cuda.BufferReader, pool: lib.MemoryManager | None = None
) -> lib.Message: ...
def read_record_batch(
    buffer: lib.Buffer,
    object: lib.Schema,
    *,
    dictionary_memo: lib.DictionaryMemo | None = None,
    pool: lib.MemoryPool | None = None,
) -> lib.RecordBatch: ...



from typing import IO, Literal

from _typeshed import StrPath

from .lib import MemoryPool, Schema, Table, _Weakrefable

class ReadOptions(_Weakrefable):
    use_threads: bool
    block_size: int
    def __init__(self, use_threads: bool | None = None, block_size: int | None = None): ...
    def equals(self, other: ReadOptions) -> bool: ...

class ParseOptions(_Weakrefable):
    explicit_schema: Schema
    newlines_in_values: bool
    unexpected_field_behavior: Literal["ignore", "error", "infer"]
    def __init__(
        self,
        explicit_schema: Schema | None = None,
        newlines_in_values: bool | None = None,
        unexpected_field_behavior: Literal["ignore", "error", "infer"] = "infer",
    ): ...
    def equals(self, other: ParseOptions) -> bool: ...

def read_json(
    input_file: StrPath | IO,
    read_options: ReadOptions | None = None,
    parse_options: ParseOptions | None = None,
    memory_pool: MemoryPool | None = None,
) -> Table: ...



from pyarrow._cuda import (
    BufferReader,
    BufferWriter,
    Context,
    CudaBuffer,
    HostBuffer,
    IpcMemHandle,
    new_host_buffer,
    read_message,
    read_record_batch,
    serialize_record_batch,
)

__all__ = [
    "BufferReader",
    "BufferWriter",
    "Context",
    "CudaBuffer",
    "HostBuffer",
    "IpcMemHandle",
    "new_host_buffer",
    "read_message",
    "read_record_batch",
    "serialize_record_batch",
]



from pyarrow.lib import _Weakrefable

class MemoryPool(_Weakrefable):
    def release_unused(self) -> None: ...
    def bytes_allocated(self) -> int: ...
    def max_memory(self) -> int | None: ...
    @property
    def backend_name(self) -> str: ...

class LoggingMemoryPool(MemoryPool): ...
class ProxyMemoryPool(MemoryPool): ...

def default_memory_pool() -> MemoryPool: ...
def proxy_memory_pool(parent: MemoryPool) -> ProxyMemoryPool: ...
def logging_memory_pool(parent: MemoryPool) -> LoggingMemoryPool: ...
def system_memory_pool() -> MemoryPool: ...
def jemalloc_memory_pool() -> MemoryPool: ...
def mimalloc_memory_pool() -> MemoryPool: ...
def set_memory_pool(pool: MemoryPool) -> None: ...
def log_memory_allocations(enable: bool = True) -> None: ...
def total_allocated_bytes() -> int: ...
def jemalloc_set_decay_ms(decay_ms: int) -> None: ...
def supported_memory_backends() -> list[str]: ...

__all__ = [
    "MemoryPool",
    "LoggingMemoryPool",
    "ProxyMemoryPool",
    "default_memory_pool",
    "proxy_memory_pool",
    "logging_memory_pool",
    "system_memory_pool",
    "jemalloc_memory_pool",
    "mimalloc_memory_pool",
    "set_memory_pool",
    "log_memory_allocations",
    "total_allocated_bytes",
    "jemalloc_set_decay_ms",
    "supported_memory_backends",
]



from pyarrow._fs import (  # noqa
    FileSelector,
    FileType,
    FileInfo,
    FileSystem,
    LocalFileSystem,
    SubTreeFileSystem,
    _MockFileSystem,
    FileSystemHandler,
    PyFileSystem,
    SupportedFileSystem,
)
from pyarrow._azurefs import AzureFileSystem
from pyarrow._hdfs import HadoopFileSystem
from pyarrow._gcsfs import GcsFileSystem
from pyarrow._s3fs import (  # noqa
    AwsDefaultS3RetryStrategy,
    AwsStandardS3RetryStrategy,
    S3FileSystem,
    S3LogLevel,
    S3RetryStrategy,
    ensure_s3_initialized,
    finalize_s3,
    ensure_s3_finalized,
    initialize_s3,
    resolve_s3_region,
)

FileStats = FileInfo

def copy_files(
    source: str,
    destination: str,
    source_filesystem: SupportedFileSystem | None = None,
    destination_filesystem: SupportedFileSystem | None = None,
    *,
    chunk_size: int = 1024 * 1024,
    use_threads: bool = True,
) -> None: ...

class FSSpecHandler(FileSystemHandler):  # type: ignore[misc]
    fs: SupportedFileSystem
    def __init__(self, fs: SupportedFileSystem) -> None: ...

__all__ = [
    # _fs
    "FileSelector",
    "FileType",
    "FileInfo",
    "FileSystem",
    "LocalFileSystem",
    "SubTreeFileSystem",
    "_MockFileSystem",
    "FileSystemHandler",
    "PyFileSystem",
    # _azurefs
    "AzureFileSystem",
    # _hdfs
    "HadoopFileSystem",
    # _gcsfs
    "GcsFileSystem",
    # _s3fs
    "AwsDefaultS3RetryStrategy",
    "AwsStandardS3RetryStrategy",
    "S3FileSystem",
    "S3LogLevel",
    "S3RetryStrategy",
    "ensure_s3_initialized",
    "finalize_s3",
    "ensure_s3_finalized",
    "initialize_s3",
    "resolve_s3_region",
    # fs
    "FileStats",
    "copy_files",
    "FSSpecHandler",
]



from pyarrow._substrait import (
    BoundExpressions,
    deserialize_expressions,
    get_supported_functions,
    run_query,
    serialize_expressions,
)

__all__ = [
    "BoundExpressions",
    "get_supported_functions",
    "run_query",
    "deserialize_expressions",
    "serialize_expressions",
]



# ruff: noqa: F401, I001, E402
__version__: str

import pyarrow.lib as _lib

_gc_enabled: bool

from pyarrow.lib import (
    BuildInfo,
    RuntimeInfo,
    set_timezone_db_path,
    MonthDayNano,
    VersionInfo,
    cpp_build_info,
    cpp_version,
    cpp_version_info,
    runtime_info,
    cpu_count,
    set_cpu_count,
    enable_signal_handlers,
    io_thread_count,
    set_io_thread_count,
)

def show_versions() -> None: ...
def show_info() -> None: ...
def _module_is_available(module: str) -> bool: ...
def _filesystem_is_available(fs: str) -> bool: ...

from pyarrow.lib import (
    null,
    bool_,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
    time32,
    time64,
    timestamp,
    date32,
    date64,
    duration,
    month_day_nano_interval,
    float16,
    float32,
    float64,
    binary,
    string,
    utf8,
    binary_view,
    string_view,
    large_binary,
    large_string,
    large_utf8,
    decimal128,
    decimal256,
    list_,
    large_list,
    list_view,
    large_list_view,
    map_,
    struct,
    union,
    sparse_union,
    dense_union,
    dictionary,
    run_end_encoded,
    fixed_shape_tensor,
    field,
    type_for_alias,
    DataType,
    DictionaryType,
    StructType,
    ListType,
    LargeListType,
    FixedSizeListType,
    ListViewType,
    LargeListViewType,
    MapType,
    UnionType,
    SparseUnionType,
    DenseUnionType,
    TimestampType,
    Time32Type,
    Time64Type,
    DurationType,
    FixedSizeBinaryType,
    Decimal128Type,
    Decimal256Type,
    BaseExtensionType,
    ExtensionType,
    RunEndEncodedType,
    FixedShapeTensorType,
    PyExtensionType,
    UnknownExtensionType,
    register_extension_type,
    unregister_extension_type,
    DictionaryMemo,
    KeyValueMetadata,
    Field,
    Schema,
    schema,
    unify_schemas,
    Array,
    Tensor,
    array,
    chunked_array,
    record_batch,
    nulls,
    repeat,
    SparseCOOTensor,
    SparseCSRMatrix,
    SparseCSCMatrix,
    SparseCSFTensor,
    infer_type,
    from_numpy_dtype,
    NullArray,
    NumericArray,
    IntegerArray,
    FloatingPointArray,
    BooleanArray,
    Int8Array,
    UInt8Array,
    Int16Array,
    UInt16Array,
    Int32Array,
    UInt32Array,
    Int64Array,
    UInt64Array,
    HalfFloatArray,
    FloatArray,
    DoubleArray,
    ListArray,
    LargeListArray,
    FixedSizeListArray,
    ListViewArray,
    LargeListViewArray,
    MapArray,
    UnionArray,
    BinaryArray,
    StringArray,
    LargeBinaryArray,
    LargeStringArray,
    BinaryViewArray,
    StringViewArray,
    FixedSizeBinaryArray,
    DictionaryArray,
    Date32Array,
    Date64Array,
    TimestampArray,
    Time32Array,
    Time64Array,
    DurationArray,
    MonthDayNanoIntervalArray,
    Decimal128Array,
    Decimal256Array,
    StructArray,
    ExtensionArray,
    RunEndEncodedArray,
    FixedShapeTensorArray,
    scalar,
    NA,
    _NULL as NULL,
    Scalar,
    NullScalar,
    BooleanScalar,
    Int8Scalar,
    Int16Scalar,
    Int32Scalar,
    Int64Scalar,
    UInt8Scalar,
    UInt16Scalar,
    UInt32Scalar,
    UInt64Scalar,
    HalfFloatScalar,
    FloatScalar,
    DoubleScalar,
    Decimal128Scalar,
    Decimal256Scalar,
    ListScalar,
    LargeListScalar,
    FixedSizeListScalar,
    ListViewScalar,
    LargeListViewScalar,
    Date32Scalar,
    Date64Scalar,
    Time32Scalar,
    Time64Scalar,
    TimestampScalar,
    DurationScalar,
    MonthDayNanoIntervalScalar,
    BinaryScalar,
    LargeBinaryScalar,
    BinaryViewScalar,
    StringScalar,
    LargeStringScalar,
    StringViewScalar,
    FixedSizeBinaryScalar,
    DictionaryScalar,
    MapScalar,
    StructScalar,
    UnionScalar,
    RunEndEncodedScalar,
    ExtensionScalar,
)

# Buffers, allocation
from pyarrow.lib import DeviceAllocationType, Device, MemoryManager, default_cpu_memory_manager

from pyarrow.lib import (
    Buffer,
    ResizableBuffer,
    foreign_buffer,
    py_buffer,
    Codec,
    compress,
    decompress,
    allocate_buffer,
)

from pyarrow.lib import (
    MemoryPool,
    LoggingMemoryPool,
    ProxyMemoryPool,
    total_allocated_bytes,
    set_memory_pool,
    default_memory_pool,
    system_memory_pool,
    jemalloc_memory_pool,
    mimalloc_memory_pool,
    logging_memory_pool,
    proxy_memory_pool,
    log_memory_allocations,
    jemalloc_set_decay_ms,
    supported_memory_backends,
)

# I/O
from pyarrow.lib import (
    NativeFile,
    PythonFile,
    BufferedInputStream,
    BufferedOutputStream,
    CacheOptions,
    CompressedInputStream,
    CompressedOutputStream,
    TransformInputStream,
    transcoding_input_stream,
    FixedSizeBufferWriter,
    BufferReader,
    BufferOutputStream,
    OSFile,
    MemoryMappedFile,
    memory_map,
    create_memory_map,
    MockOutputStream,
    input_stream,
    output_stream,
    have_libhdfs,
)

from pyarrow.lib import (
    ChunkedArray,
    RecordBatch,
    Table,
    table,
    concat_arrays,
    concat_tables,
    TableGroupBy,
    RecordBatchReader,
)

# Exceptions
from pyarrow.lib import (
    ArrowCancelled,
    ArrowCapacityError,
    ArrowException,
    ArrowKeyError,
    ArrowIndexError,
    ArrowInvalid,
    ArrowIOError,
    ArrowMemoryError,
    ArrowNotImplementedError,
    ArrowTypeError,
    ArrowSerializationError,
)

from pyarrow.ipc import serialize_pandas, deserialize_pandas
import pyarrow.ipc as ipc

import pyarrow.types as types

# ----------------------------------------------------------------------
# Deprecations

from pyarrow.util import _deprecate_api, _deprecate_class

from pyarrow.ipc import (
    Message,
    MessageReader,
    MetadataVersion,
    RecordBatchFileReader,
    RecordBatchFileWriter,
    RecordBatchStreamReader,
    RecordBatchStreamWriter,
)

# ----------------------------------------------------------------------
# Returning absolute path to the pyarrow include directory (if bundled, e.g. in
# wheels)
def get_include() -> str: ...
def _get_pkg_config_executable() -> str: ...
def _has_pkg_config(pkgname: str) -> bool: ...
def _read_pkg_config_variable(pkgname: str, cli_args: list[str]) -> str: ...
def get_libraries() -> list[str]: ...
def create_library_symlinks() -> None: ...
def get_library_dirs() -> list[str]: ...

__all__ = [
    "__version__",
    "_lib",
    "_gc_enabled",
    "BuildInfo",
    "RuntimeInfo",
    "set_timezone_db_path",
    "MonthDayNano",
    "VersionInfo",
    "cpp_build_info",
    "cpp_version",
    "cpp_version_info",
    "runtime_info",
    "cpu_count",
    "set_cpu_count",
    "enable_signal_handlers",
    "io_thread_count",
    "set_io_thread_count",
    "show_versions",
    "show_info",
    "_module_is_available",
    "_filesystem_is_available",
    "null",
    "bool_",
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "time32",
    "time64",
    "timestamp",
    "date32",
    "date64",
    "duration",
    "month_day_nano_interval",
    "float16",
    "float32",
    "float64",
    "binary",
    "string",
    "utf8",
    "binary_view",
    "string_view",
    "large_binary",
    "large_string",
    "large_utf8",
    "decimal128",
    "decimal256",
    "list_",
    "large_list",
    "list_view",
    "large_list_view",
    "map_",
    "struct",
    "union",
    "sparse_union",
    "dense_union",
    "dictionary",
    "run_end_encoded",
    "fixed_shape_tensor",
    "field",
    "type_for_alias",
    "DataType",
    "DictionaryType",
    "StructType",
    "ListType",
    "LargeListType",
    "FixedSizeListType",
    "ListViewType",
    "LargeListViewType",
    "MapType",
    "UnionType",
    "SparseUnionType",
    "DenseUnionType",
    "TimestampType",
    "Time32Type",
    "Time64Type",
    "DurationType",
    "FixedSizeBinaryType",
    "Decimal128Type",
    "Decimal256Type",
    "BaseExtensionType",
    "ExtensionType",
    "RunEndEncodedType",
    "FixedShapeTensorType",
    "PyExtensionType",
    "UnknownExtensionType",
    "register_extension_type",
    "unregister_extension_type",
    "DictionaryMemo",
    "KeyValueMetadata",
    "Field",
    "Schema",
    "schema",
    "unify_schemas",
    "Array",
    "Tensor",
    "array",
    "chunked_array",
    "record_batch",
    "nulls",
    "repeat",
    "SparseCOOTensor",
    "SparseCSRMatrix",
    "SparseCSCMatrix",
    "SparseCSFTensor",
    "infer_type",
    "from_numpy_dtype",
    "NullArray",
    "NumericArray",
    "IntegerArray",
    "FloatingPointArray",
    "BooleanArray",
    "Int8Array",
    "UInt8Array",
    "Int16Array",
    "UInt16Array",
    "Int32Array",
    "UInt32Array",
    "Int64Array",
    "UInt64Array",
    "HalfFloatArray",
    "FloatArray",
    "DoubleArray",
    "ListArray",
    "LargeListArray",
    "FixedSizeListArray",
    "ListViewArray",
    "LargeListViewArray",
    "MapArray",
    "UnionArray",
    "BinaryArray",
    "StringArray",
    "LargeBinaryArray",
    "LargeStringArray",
    "BinaryViewArray",
    "StringViewArray",
    "FixedSizeBinaryArray",
    "DictionaryArray",
    "Date32Array",
    "Date64Array",
    "TimestampArray",
    "Time32Array",
    "Time64Array",
    "DurationArray",
    "MonthDayNanoIntervalArray",
    "Decimal128Array",
    "Decimal256Array",
    "StructArray",
    "ExtensionArray",
    "RunEndEncodedArray",
    "FixedShapeTensorArray",
    "scalar",
    "NA",
    "NULL",
    "Scalar",
    "NullScalar",
    "BooleanScalar",
    "Int8Scalar",
    "Int16Scalar",
    "Int32Scalar",
    "Int64Scalar",
    "UInt8Scalar",
    "UInt16Scalar",
    "UInt32Scalar",
    "UInt64Scalar",
    "HalfFloatScalar",
    "FloatScalar",
    "DoubleScalar",
    "Decimal128Scalar",
    "Decimal256Scalar",
    "ListScalar",
    "LargeListScalar",
    "FixedSizeListScalar",
    "ListViewScalar",
    "LargeListViewScalar",
    "Date32Scalar",
    "Date64Scalar",
    "Time32Scalar",
    "Time64Scalar",
    "TimestampScalar",
    "DurationScalar",
    "MonthDayNanoIntervalScalar",
    "BinaryScalar",
    "LargeBinaryScalar",
    "BinaryViewScalar",
    "StringScalar",
    "LargeStringScalar",
    "StringViewScalar",
    "FixedSizeBinaryScalar",
    "DictionaryScalar",
    "MapScalar",
    "StructScalar",
    "UnionScalar",
    "RunEndEncodedScalar",
    "ExtensionScalar",
    "DeviceAllocationType",
    "Device",
    "MemoryManager",
    "default_cpu_memory_manager",
    "Buffer",
    "ResizableBuffer",
    "foreign_buffer",
    "py_buffer",
    "Codec",
    "compress",
    "decompress",
    "allocate_buffer",
    "MemoryPool",
    "LoggingMemoryPool",
    "ProxyMemoryPool",
    "total_allocated_bytes",
    "set_memory_pool",
    "default_memory_pool",
    "system_memory_pool",
    "jemalloc_memory_pool",
    "mimalloc_memory_pool",
    "logging_memory_pool",
    "proxy_memory_pool",
    "log_memory_allocations",
    "jemalloc_set_decay_ms",
    "supported_memory_backends",
    "NativeFile",
    "PythonFile",
    "BufferedInputStream",
    "BufferedOutputStream",
    "CacheOptions",
    "CompressedInputStream",
    "CompressedOutputStream",
    "TransformInputStream",
    "transcoding_input_stream",
    "FixedSizeBufferWriter",
    "BufferReader",
    "BufferOutputStream",
    "OSFile",
    "MemoryMappedFile",
    "memory_map",
    "create_memory_map",
    "MockOutputStream",
    "input_stream",
    "output_stream",
    "have_libhdfs",
    "ChunkedArray",
    "RecordBatch",
    "Table",
    "table",
    "concat_arrays",
    "concat_tables",
    "TableGroupBy",
    "RecordBatchReader",
    "ArrowCancelled",
    "ArrowCapacityError",
    "ArrowException",
    "ArrowKeyError",
    "ArrowIndexError",
    "ArrowInvalid",
    "ArrowIOError",
    "ArrowMemoryError",
    "ArrowNotImplementedError",
    "ArrowTypeError",
    "ArrowSerializationError",
    "serialize_pandas",
    "deserialize_pandas",
    "ipc",
    "types",
    "_deprecate_api",
    "_deprecate_class",
    "Message",
    "MessageReader",
    "MetadataVersion",
    "RecordBatchFileReader",
    "RecordBatchFileWriter",
    "RecordBatchStreamReader",
    "RecordBatchStreamWriter",
    "get_include",
    "_get_pkg_config_executable",
    "_has_pkg_config",
    "_read_pkg_config_variable",
    "get_libraries",
    "create_library_symlinks",
    "get_library_dirs",
]



import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias
from typing import Literal

from . import lib
from .compute import Expression, FunctionOptions

_StrOrExpr: TypeAlias = str | Expression

class Declaration(lib._Weakrefable):
    def __init__(
        self,
        factory_name: str,
        options: ExecNodeOptions,
        inputs: list[Declaration] | None = None,
    ) -> None: ...
    @classmethod
    def from_sequence(cls, decls: list[Declaration]) -> Self: ...
    def to_reader(self, use_threads: bool = True) -> lib.RecordBatchReader: ...
    def to_table(self, use_threads: bool = True) -> lib.Table: ...

class ExecNodeOptions(lib._Weakrefable): ...

class TableSourceNodeOptions(ExecNodeOptions):
    def __init__(self, table: lib.Table) -> None: ...

class FilterNodeOptions(ExecNodeOptions):
    def __init__(self, filter_expression: Expression) -> None: ...

class ProjectNodeOptions(ExecNodeOptions):
    def __init__(self, expressions: list[Expression], names: list[str] | None = None) -> None: ...

class AggregateNodeOptions(ExecNodeOptions):
    def __init__(
        self,
        aggregates: list[tuple[list[str], str, FunctionOptions, str]],
        keys: list[_StrOrExpr] | None = None,
    ) -> None: ...

class OrderByNodeOptions(ExecNodeOptions):
    def __init__(
        self,
        sort_keys: tuple[tuple[str, Literal["ascending", "descending"]], ...] = (),
        *,
        null_placement: Literal["at_start", "at_end"] = "at_end",
    ) -> None: ...

class HashJoinNodeOptions(ExecNodeOptions):
    def __init__(
        self,
        join_type: Literal[
            "left semi",
            "right semi",
            "left anti",
            "right anti",
            "inner",
            "left outer",
            "right outer",
            "full outer",
        ],
        left_keys: _StrOrExpr | list[_StrOrExpr],
        right_keys: _StrOrExpr | list[_StrOrExpr],
        left_output: list[_StrOrExpr] | None = None,
        right_output: list[_StrOrExpr] | None = None,
        output_suffix_for_left: str = "",
        output_suffix_for_right: str = "",
    ) -> None: ...

class AsofJoinNodeOptions(ExecNodeOptions):
    def __init__(
        self,
        left_on: _StrOrExpr,
        left_by: _StrOrExpr | list[_StrOrExpr],
        right_on: _StrOrExpr,
        right_by: _StrOrExpr | list[_StrOrExpr],
        tolerance: int,
    ) -> None: ...



# mypy: disable-error-code="misc,type-var,var-annotated"
# ruff: noqa: I001
from typing import Literal, TypeAlias, TypeVar, overload, Any, Iterable, ParamSpec, Sequence
from collections.abc import Callable

# Option classes
from pyarrow._compute import ArraySortOptions as ArraySortOptions
from pyarrow._compute import AssumeTimezoneOptions as AssumeTimezoneOptions
from pyarrow._compute import CastOptions as CastOptions
from pyarrow._compute import CountOptions as CountOptions
from pyarrow._compute import CumulativeOptions as CumulativeOptions
from pyarrow._compute import CumulativeSumOptions as CumulativeSumOptions
from pyarrow._compute import DayOfWeekOptions as DayOfWeekOptions
from pyarrow._compute import DictionaryEncodeOptions as DictionaryEncodeOptions
from pyarrow._compute import ElementWiseAggregateOptions as ElementWiseAggregateOptions

# Expressions
from pyarrow._compute import Expression as Expression
from pyarrow._compute import ExtractRegexOptions as ExtractRegexOptions
from pyarrow._compute import FilterOptions as FilterOptions
from pyarrow._compute import Function as Function
from pyarrow._compute import FunctionOptions as FunctionOptions
from pyarrow._compute import FunctionRegistry as FunctionRegistry
from pyarrow._compute import HashAggregateFunction as HashAggregateFunction
from pyarrow._compute import HashAggregateKernel as HashAggregateKernel
from pyarrow._compute import IndexOptions as IndexOptions
from pyarrow._compute import JoinOptions as JoinOptions
from pyarrow._compute import Kernel as Kernel
from pyarrow._compute import ListFlattenOptions as ListFlattenOptions
from pyarrow._compute import ListSliceOptions as ListSliceOptions
from pyarrow._compute import MakeStructOptions as MakeStructOptions
from pyarrow._compute import MapLookupOptions as MapLookupOptions
from pyarrow._compute import MatchSubstringOptions as MatchSubstringOptions
from pyarrow._compute import ModeOptions as ModeOptions
from pyarrow._compute import NullOptions as NullOptions
from pyarrow._compute import PadOptions as PadOptions
from pyarrow._compute import PairwiseOptions as PairwiseOptions
from pyarrow._compute import PartitionNthOptions as PartitionNthOptions
from pyarrow._compute import QuantileOptions as QuantileOptions
from pyarrow._compute import RandomOptions as RandomOptions
from pyarrow._compute import RankOptions as RankOptions
from pyarrow._compute import ReplaceSliceOptions as ReplaceSliceOptions
from pyarrow._compute import ReplaceSubstringOptions as ReplaceSubstringOptions
from pyarrow._compute import RoundBinaryOptions as RoundBinaryOptions
from pyarrow._compute import RoundOptions as RoundOptions
from pyarrow._compute import RoundTemporalOptions as RoundTemporalOptions
from pyarrow._compute import RoundToMultipleOptions as RoundToMultipleOptions
from pyarrow._compute import RunEndEncodeOptions as RunEndEncodeOptions
from pyarrow._compute import ScalarAggregateFunction as ScalarAggregateFunction
from pyarrow._compute import ScalarAggregateKernel as ScalarAggregateKernel
from pyarrow._compute import ScalarAggregateOptions as ScalarAggregateOptions
from pyarrow._compute import ScalarFunction as ScalarFunction
from pyarrow._compute import ScalarKernel as ScalarKernel
from pyarrow._compute import SelectKOptions as SelectKOptions
from pyarrow._compute import SetLookupOptions as SetLookupOptions
from pyarrow._compute import SliceOptions as SliceOptions
from pyarrow._compute import SortOptions as SortOptions
from pyarrow._compute import SplitOptions as SplitOptions
from pyarrow._compute import SplitPatternOptions as SplitPatternOptions
from pyarrow._compute import StrftimeOptions as StrftimeOptions
from pyarrow._compute import StrptimeOptions as StrptimeOptions
from pyarrow._compute import StructFieldOptions as StructFieldOptions
from pyarrow._compute import TakeOptions as TakeOptions
from pyarrow._compute import TDigestOptions as TDigestOptions
from pyarrow._compute import TrimOptions as TrimOptions
from pyarrow._compute import UdfContext as UdfContext
from pyarrow._compute import Utf8NormalizeOptions as Utf8NormalizeOptions
from pyarrow._compute import VarianceOptions as VarianceOptions
from pyarrow._compute import VectorFunction as VectorFunction
from pyarrow._compute import VectorKernel as VectorKernel
from pyarrow._compute import WeekOptions as WeekOptions

# Functions
from pyarrow._compute import call_function as call_function

# Udf
from pyarrow._compute import call_tabular_function as call_tabular_function
from pyarrow._compute import function_registry as function_registry
from pyarrow._compute import get_function as get_function
from pyarrow._compute import list_functions as list_functions
from pyarrow._compute import register_aggregate_function as register_aggregate_function
from pyarrow._compute import register_scalar_function as register_scalar_function
from pyarrow._compute import register_tabular_function as register_tabular_function
from pyarrow._compute import register_vector_function as register_vector_function

from . import lib

_P = ParamSpec("_P")
_R = TypeVar("_R")

def field(*name_or_index: str | tuple[str, ...] | int) -> Expression: ...
def scalar(value: bool | float | str) -> Expression: ...
def _clone_signature(f: Callable[_P, _R]) -> Callable[_P, _R]: ...

# ============= compute functions =============
_DataTypeT = TypeVar("_DataTypeT", bound=lib.DataType)
NumericScalar: TypeAlias = (
    lib.Scalar[lib.Int8Type]
    | lib.Scalar[lib.Int16Type]
    | lib.Scalar[lib.Int32Type]
    | lib.Scalar[lib.Int64Type]
    | lib.Scalar[lib.Uint8Type]
    | lib.Scalar[lib.Uint16Type]
    | lib.Scalar[lib.Uint32Type]
    | lib.Scalar[lib.Uint64Type]
    | lib.Scalar[lib.Float16Type]
    | lib.Scalar[lib.Float32Type]
    | lib.Scalar[lib.Float64Type]
    | lib.Scalar[lib.Decimal128Type]
    | lib.Scalar[lib.Decimal256Type]
)
BinaryScalar: TypeAlias = (
    lib.Scalar[lib.BinaryType]
    | lib.Scalar[lib.LargeBinaryType]
    | lib.Scalar[lib.FixedSizeBinaryType]
)
StringScalar: TypeAlias = lib.Scalar[lib.StringType] | lib.Scalar[lib.LargeStringType]
ListScalar: TypeAlias = (
    lib.ListScalar[_DataTypeT]
    | lib.LargeListScalar[_DataTypeT]
    | lib.ListViewScalar[_DataTypeT]
    | lib.LargeListViewScalar[_DataTypeT]
    | lib.FixedSizeListScalar[_DataTypeT, Any]
)
TemporalScalar: TypeAlias = (
    lib.Date32Scalar
    | lib.Date64Scalar
    | lib.Time32Scalar
    | lib.Time64Scalar
    | lib.TimestampScalar
    | lib.DurationScalar
    | lib.MonthDayNanoIntervalScalar
)
_NumericScalarT = TypeVar("_NumericScalarT", bound=NumericScalar)
NumericOrDurationScalar: TypeAlias = NumericScalar | lib.DurationScalar
_NumericOrDurationT = TypeVar("_NumericOrDurationT", bound=NumericOrDurationScalar)
NumericOrTemporalScalar: TypeAlias = NumericScalar | TemporalScalar
_NumericOrTemporalT = TypeVar("_NumericOrTemporalT", bound=NumericOrTemporalScalar)
NumericArray: TypeAlias = lib.NumericArray[_ScalarT] | lib.ChunkedArray[_ScalarT]
_NumericArrayT = TypeVar("_NumericArrayT", bound=lib.NumericArray)
NumericOrDurationArray: TypeAlias = (
    lib.NumericArray | lib.Array[lib.DurationScalar] | lib.ChunkedArray
)
_NumericOrDurationArrayT = TypeVar("_NumericOrDurationArrayT", bound=NumericOrDurationArray)
NumericOrTemporalArray: TypeAlias = (
    lib.NumericArray | lib.Array[TemporalScalar] | lib.ChunkedArray[TemporalScalar]
)
_NumericOrTemporalArrayT = TypeVar("_NumericOrTemporalArrayT", bound=NumericOrTemporalArray)
BooleanArray: TypeAlias = lib.BooleanArray | lib.ChunkedArray[lib.BooleanScalar]
FloatScalar: TypeAlias = lib.Scalar[lib.Float32Type] | lib.Scalar[lib.Float64Type]
DecimalScalar: TypeAlias = lib.Scalar[lib.Decimal128Type] | lib.Scalar[lib.Decimal256Type]
_FloatScalarT = TypeVar("_FloatScalarT", bound=FloatScalar)
FloatArray: TypeAlias = (
    lib.NumericArray[lib.FloatScalar]
    | lib.NumericArray[lib.DoubleScalar]
    | lib.ChunkedArray[lib.FloatScalar]
    | lib.ChunkedArray[lib.DoubleScalar]
)

_FloatArrayT = TypeVar("_FloatArrayT", bound=FloatArray)
_StringScalarT = TypeVar("_StringScalarT", bound=StringScalar)
StringArray: TypeAlias = (
    lib.StringArray
    | lib.LargeStringArray
    | lib.ChunkedArray[lib.StringScalar]
    | lib.ChunkedArray[lib.LargeStringScalar]
)
_StringArrayT = TypeVar("_StringArrayT", bound=StringArray)
_BinaryScalarT = TypeVar("_BinaryScalarT", bound=BinaryScalar)
BinaryArray: TypeAlias = (
    lib.BinaryArray
    | lib.LargeBinaryArray
    | lib.ChunkedArray[lib.BinaryScalar]
    | lib.ChunkedArray[lib.LargeBinaryScalar]
)
_BinaryArrayT = TypeVar("_BinaryArrayT", bound=BinaryArray)
StringOrBinaryScalar: TypeAlias = StringScalar | BinaryScalar
_StringOrBinaryScalarT = TypeVar("_StringOrBinaryScalarT", bound=StringOrBinaryScalar)
StringOrBinaryArray: TypeAlias = StringArray | BinaryArray
_StringOrBinaryArrayT = TypeVar("_StringOrBinaryArrayT", bound=StringOrBinaryArray)
_TemporalScalarT = TypeVar("_TemporalScalarT", bound=TemporalScalar)
TemporalArray: TypeAlias = (
    lib.Date32Array
    | lib.Date64Array
    | lib.Time32Array
    | lib.Time64Array
    | lib.TimestampArray
    | lib.DurationArray
    | lib.MonthDayNanoIntervalArray
    | lib.ChunkedArray[lib.Date32Scalar]
    | lib.ChunkedArray[lib.Date64Scalar]
    | lib.ChunkedArray[lib.Time32Scalar]
    | lib.ChunkedArray[lib.Time64Scalar]
    | lib.ChunkedArray[lib.DurationScalar]
    | lib.ChunkedArray[lib.MonthDayNanoIntervalScalar]
)
_TemporalArrayT = TypeVar("_TemporalArrayT", bound=TemporalArray)
_ScalarT = TypeVar("_ScalarT", bound=lib.Scalar)
_ArrayT = TypeVar("_ArrayT", bound=lib.Array)
_ScalarOrArrayT = TypeVar("_ScalarOrArrayT", bound=lib.Array | lib.Scalar)
# =============================== 1. Aggregation ===============================

# ========================= 1.1 functions =========================

def all(
    array: lib.BooleanScalar | BooleanArray,
    /,
    *,
    skip_nulls: bool = True,
    min_count: int = 1,
    options: ScalarAggregateOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanScalar: ...

any = _clone_signature(all)

def approximate_median(
    array: NumericScalar | NumericArray,
    /,
    *,
    skip_nulls: bool = True,
    min_count: int = 1,
    options: ScalarAggregateOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.DoubleScalar: ...
def count(
    array: lib.Array | lib.ChunkedArray,
    /,
    mode: Literal["only_valid", "only_null", "all"] = "only_valid",
    *,
    options: CountOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Scalar: ...
def count_distinct(
    array: lib.Array | lib.ChunkedArray,
    /,
    mode: Literal["only_valid", "only_null", "all"] = "only_valid",
    *,
    options: CountOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Scalar: ...
def first(
    array: lib.Array[_ScalarT] | lib.ChunkedArray[_ScalarT],
    /,
    *,
    skip_nulls: bool = True,
    min_count: int = 1,
    options: ScalarAggregateOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _ScalarT: ...
def first_last(
    array: lib.Array | lib.ChunkedArray,
    /,
    *,
    skip_nulls: bool = True,
    min_count: int = 1,
    options: ScalarAggregateOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.StructScalar: ...
def index(
    data: lib.Array | lib.ChunkedArray,
    value,
    start: int | None = None,
    end: int | None = None,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Scalar: ...

last = _clone_signature(first)
max = _clone_signature(first)
min = _clone_signature(first)
min_max = _clone_signature(first_last)

@overload
def mean(
    array: FloatScalar | FloatArray,
    /,
    *,
    skip_nulls: bool = True,
    min_count: int = 1,
    options: ScalarAggregateOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.DoubleScalar: ...
@overload
def mean(
    array: lib.NumericArray[lib.Decimal128Scalar]
    | lib.ChunkedArray[lib.Decimal128Scalar]
    | lib.Decimal128Scalar,
    /,
    *,
    skip_nulls: bool = True,
    min_count: int = 1,
    options: ScalarAggregateOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Decimal128Scalar: ...
@overload
def mean(
    array: lib.NumericArray[lib.Decimal256Scalar]
    | lib.ChunkedArray[lib.Decimal256Scalar]
    | lib.Decimal256Scalar,
    /,
    *,
    skip_nulls: bool = True,
    min_count: int = 1,
    options: ScalarAggregateOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Decimal256Scalar: ...
def mode(
    array: NumericScalar | NumericArray,
    /,
    n: int = 1,
    *,
    skip_nulls: bool = True,
    min_count: int = 0,
    options: ModeOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.StructArray: ...
def product(
    array: _ScalarT | lib.NumericArray[_ScalarT],
    /,
    *,
    skip_nulls=True,
    min_count=1,
    options=None,
    memory_pool: lib.MemoryPool | None = None,
) -> _ScalarT: ...
def quantile(
    array: NumericScalar | NumericArray,
    /,
    q: float = 0.5,
    *,
    interpolation: Literal["linear", "lower", "higher", "nearest", "midpoint"] = "linear",
    skip_nulls: bool = True,
    min_count: int = 0,
    options: QuantileOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.DoubleArray: ...
def stddev(
    array: NumericScalar | NumericArray,
    /,
    *,
    ddof: float = 0,
    skip_nulls: bool = True,
    min_count: int = 0,
    options: VarianceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.DoubleScalar: ...
def sum(
    array: _NumericScalarT | NumericArray[_NumericScalarT],
    /,
    *,
    skip_nulls: bool = True,
    min_count: int = 1,
    options: ScalarAggregateOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericScalarT: ...
def tdigest(
    array: NumericScalar | NumericArray,
    /,
    q: float = 0.5,
    *,
    delta: int = 100,
    buffer_size: int = 500,
    skip_nulls: bool = True,
    min_count: int = 0,
    options: TDigestOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.DoubleArray: ...
def variance(
    array: NumericScalar | NumericArray,
    /,
    *,
    ddof: int = 0,
    skip_nulls: bool = True,
    min_count: int = 0,
    options: VarianceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.DoubleScalar: ...

# ========================= 2. Element-wise (“scalar”) functions =========================

# ========================= 2.1 Arithmetic =========================
@overload
def abs(
    x: _NumericOrDurationT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _NumericOrDurationT: ...
@overload
def abs(
    x: _NumericOrDurationArrayT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _NumericOrDurationArrayT: ...
@overload
def abs(x: Expression, /, *, memory_pool: lib.MemoryPool | None = None) -> Expression: ...

abs_checked = _clone_signature(abs)

@overload
def add(
    x: _NumericOrTemporalT, y: _NumericOrTemporalT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _NumericOrTemporalT: ...
@overload
def add(
    x: _NumericOrTemporalArrayT,
    y: _NumericOrTemporalArrayT,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericOrTemporalArrayT: ...
@overload
def add(
    x: NumericScalar, y: NumericScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> NumericScalar: ...
@overload
def add(
    x: TemporalScalar, y: TemporalScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> TemporalScalar: ...
@overload
def add(
    x: NumericOrTemporalArray | NumericOrTemporalScalar,
    y: NumericOrTemporalArray | NumericOrTemporalScalar,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> NumericOrTemporalArray: ...
@overload
def add(
    x: Expression | Any, y: Expression | Any, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...

add_checked = _clone_signature(add)

@overload
def divide(
    dividend: NumericScalar,
    divisor: NumericScalar,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> NumericScalar: ...
@overload
def divide(
    dividend: TemporalScalar,
    divisor: TemporalScalar,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> TemporalScalar: ...
@overload
def divide(
    dividend: NumericOrTemporalArray | NumericOrTemporalScalar,
    divisor: NumericOrTemporalArray | NumericOrTemporalScalar,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> NumericArray: ...
@overload
def divide(
    dividend: Expression | Any,
    divisor: Expression | Any,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

divide_checked = _clone_signature(divide)

@overload
def exp(
    exponent: lib.FloatArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.FloatArray: ...
@overload
def exp(
    exponent: NumericArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.DoubleArray: ...
@overload
def exp(
    exponent: lib.FloatScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.FloatScalar: ...
@overload
def exp(
    exponent: NumericScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.DoubleScalar: ...
@overload
def exp(exponent: Expression, /, *, memory_pool: lib.MemoryPool | None = None) -> Expression: ...

multiply = _clone_signature(add)
multiply_checked = _clone_signature(multiply)

@overload
def negate(
    x: _NumericOrDurationT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _NumericOrDurationT: ...
@overload
def negate(
    x: _NumericOrDurationArrayT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _NumericOrDurationArrayT: ...
@overload
def negate(x: Expression, /, *, memory_pool: lib.MemoryPool | None = None) -> Expression: ...

negate_checked = _clone_signature(negate)

@overload
def power(
    base: _NumericScalarT,
    exponent: _NumericScalarT,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericScalarT: ...
@overload
def power(
    base: NumericScalar, exponent: NumericScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> NumericScalar: ...
@overload
def power(
    base: _NumericArrayT,
    exponent: _NumericArrayT,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericArrayT: ...
@overload
def power(
    base: NumericScalar | NumericArray,
    exponent: NumericScalar | NumericArray,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> NumericArray: ...
@overload
def power(
    base: Expression | Any,
    exponent: Expression | Any,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

power_checked = _clone_signature(power)

@overload
def sign(
    x: NumericOrDurationArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> (
    lib.NumericArray[lib.Int8Scalar]
    | lib.NumericArray[lib.FloatScalar]
    | lib.NumericArray[lib.DoubleScalar]
): ...
@overload
def sign(
    x: NumericOrDurationScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.Int8Scalar | lib.FloatScalar | lib.DoubleScalar: ...
@overload
def sign(x: Expression, /, *, memory_pool: lib.MemoryPool | None = None) -> Expression: ...
@overload
def sqrt(x: NumericArray, /, *, memory_pool: lib.MemoryPool | None = None) -> FloatArray: ...
@overload
def sqrt(x: NumericScalar, /, *, memory_pool: lib.MemoryPool | None = None) -> FloatScalar: ...
@overload
def sqrt(x: Expression, /, *, memory_pool: lib.MemoryPool | None = None) -> Expression: ...

sqrt_checked = _clone_signature(sqrt)

subtract = _clone_signature(add)
subtract_checked = _clone_signature(subtract)

# ========================= 2.1 Bit-wise functions =========================
@overload
def bit_wise_and(
    x: _NumericScalarT, y: _NumericScalarT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _NumericScalarT: ...
@overload
def bit_wise_and(
    x: _NumericArrayT,
    y: _NumericArrayT,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericArrayT: ...
@overload
def bit_wise_and(
    x: NumericScalar, y: NumericScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> NumericScalar: ...
@overload
def bit_wise_and(
    x: NumericArray | NumericScalar,
    y: NumericArray | NumericScalar,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> NumericArray: ...
@overload
def bit_wise_and(
    x: Expression | Any, y: Expression | Any, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...
@overload
def bit_wise_not(
    x: _NumericScalarT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _NumericScalarT: ...
@overload
def bit_wise_not(
    x: _NumericArrayT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _NumericArrayT: ...
@overload
def bit_wise_not(x: Expression, /, *, memory_pool: lib.MemoryPool | None = None) -> Expression: ...

bit_wise_or = _clone_signature(bit_wise_and)
bit_wise_xor = _clone_signature(bit_wise_and)
shift_left = _clone_signature(bit_wise_and)
shift_left_checked = _clone_signature(bit_wise_and)
shift_right = _clone_signature(bit_wise_and)
shift_right_checked = _clone_signature(bit_wise_and)

# ========================= 2.2 Rounding functions =========================
@overload
def ceil(x: _FloatScalarT, /, *, memory_pool: lib.MemoryPool | None = None) -> _FloatScalarT: ...
@overload
def ceil(x: _FloatArrayT, /, *, memory_pool: lib.MemoryPool | None = None) -> _FloatArrayT: ...
@overload
def ceil(x: Expression, /, *, memory_pool: lib.MemoryPool | None = None) -> Expression: ...

floor = _clone_signature(ceil)

@overload
def round(
    x: _NumericScalarT,
    /,
    ndigits: int = 0,
    round_mode: Literal[
        "down",
        "up",
        "towards_zero",
        "towards_infinity",
        "half_down",
        "half_up",
        "half_towards_zero",
        "half_towards_infinity",
        "half_to_even",
        "half_to_odd",
    ] = "half_to_even",
    *,
    options: RoundOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericScalarT: ...
@overload
def round(
    x: _NumericArrayT,
    /,
    ndigits: int = 0,
    round_mode: Literal[
        "down",
        "up",
        "towards_zero",
        "towards_infinity",
        "half_down",
        "half_up",
        "half_towards_zero",
        "half_towards_infinity",
        "half_to_even",
        "half_to_odd",
    ] = "half_to_even",
    *,
    options: RoundOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericArrayT: ...
@overload
def round(
    x: Expression,
    /,
    ndigits: int = 0,
    round_mode: Literal[
        "down",
        "up",
        "towards_zero",
        "towards_infinity",
        "half_down",
        "half_up",
        "half_towards_zero",
        "half_towards_infinity",
        "half_to_even",
        "half_to_odd",
    ] = "half_to_even",
    *,
    options: RoundOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def round_to_multiple(
    x: _NumericScalarT,
    /,
    multiple: int = 0,
    round_mode: Literal[
        "down",
        "up",
        "towards_zero",
        "towards_infinity",
        "half_down",
        "half_up",
        "half_towards_zero",
        "half_towards_infinity",
        "half_to_even",
        "half_to_odd",
    ] = "half_to_even",
    *,
    options: RoundToMultipleOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericScalarT: ...
@overload
def round_to_multiple(
    x: _NumericArrayT,
    /,
    multiple: int = 0,
    round_mode: Literal[
        "down",
        "up",
        "towards_zero",
        "towards_infinity",
        "half_down",
        "half_up",
        "half_towards_zero",
        "half_towards_infinity",
        "half_to_even",
        "half_to_odd",
    ] = "half_to_even",
    *,
    options: RoundToMultipleOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericArrayT: ...
@overload
def round_to_multiple(
    x: Expression,
    /,
    multiple: int = 0,
    round_mode: Literal[
        "down",
        "up",
        "towards_zero",
        "towards_infinity",
        "half_down",
        "half_up",
        "half_towards_zero",
        "half_towards_infinity",
        "half_to_even",
        "half_to_odd",
    ] = "half_to_even",
    *,
    options: RoundToMultipleOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def round_binary(
    x: _NumericScalarT,
    s: int | lib.Int8Scalar | lib.Int16Scalar | lib.Int32Scalar | lib.Int64Scalar,
    /,
    round_mode: Literal[
        "down",
        "up",
        "towards_zero",
        "towards_infinity",
        "half_down",
        "half_up",
        "half_towards_zero",
        "half_towards_infinity",
        "half_to_even",
        "half_to_odd",
    ] = "half_to_even",
    *,
    options: RoundBinaryOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericScalarT: ...
@overload
def round_binary(
    x: _NumericScalarT,
    s: Iterable,
    /,
    round_mode: Literal[
        "down",
        "up",
        "towards_zero",
        "towards_infinity",
        "half_down",
        "half_up",
        "half_towards_zero",
        "half_towards_infinity",
        "half_to_even",
        "half_to_odd",
    ] = "half_to_even",
    *,
    options: RoundBinaryOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.NumericArray[_NumericScalarT]: ...
@overload
def round_binary(
    x: _NumericArrayT,
    s: int | lib.Int8Scalar | lib.Int16Scalar | lib.Int32Scalar | lib.Int64Scalar | Iterable,
    /,
    round_mode: Literal[
        "down",
        "up",
        "towards_zero",
        "towards_infinity",
        "half_down",
        "half_up",
        "half_towards_zero",
        "half_towards_infinity",
        "half_to_even",
        "half_to_odd",
    ] = "half_to_even",
    *,
    options: RoundBinaryOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericArrayT: ...
@overload
def round_binary(
    x: Expression,
    s: Iterable,
    /,
    round_mode: Literal[
        "down",
        "up",
        "towards_zero",
        "towards_infinity",
        "half_down",
        "half_up",
        "half_towards_zero",
        "half_towards_infinity",
        "half_to_even",
        "half_to_odd",
    ] = "half_to_even",
    *,
    options: RoundBinaryOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

trunc = _clone_signature(ceil)

# ========================= 2.3 Logarithmic functions =========================
@overload
def ln(
    x: FloatScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.FloatScalar | lib.DoubleScalar: ...
@overload
def ln(
    x: FloatArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.NumericArray[lib.FloatScalar] | lib.NumericArray[lib.DoubleScalar]: ...
@overload
def ln(x: Expression, /, *, memory_pool: lib.MemoryPool | None = None) -> Expression: ...

ln_checked = _clone_signature(ln)
log10 = _clone_signature(ln)
log10_checked = _clone_signature(ln)
log1p = _clone_signature(ln)
log1p_checked = _clone_signature(ln)
log2 = _clone_signature(ln)
log2_checked = _clone_signature(ln)

@overload
def logb(
    x: FloatScalar, b: FloatScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.FloatScalar | lib.DoubleScalar: ...
@overload
def logb(
    x: FloatArray, b: FloatArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.NumericArray[lib.FloatScalar] | lib.NumericArray[lib.DoubleScalar]: ...
@overload
def logb(
    x: FloatScalar | FloatArray,
    b: FloatScalar | FloatArray,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.NumericArray[lib.FloatScalar] | lib.NumericArray[lib.DoubleScalar]: ...
@overload
def logb(
    x: Expression | Any, b: Expression | Any, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression | Any: ...

logb_checked = _clone_signature(logb)

# ========================= 2.4 Trigonometric functions =========================
acos = _clone_signature(ln)
acos_checked = _clone_signature(ln)
asin = _clone_signature(ln)
asin_checked = _clone_signature(ln)
atan = _clone_signature(ln)
cos = _clone_signature(ln)
cos_checked = _clone_signature(ln)
sin = _clone_signature(ln)
sin_checked = _clone_signature(ln)
tan = _clone_signature(ln)
tan_checked = _clone_signature(ln)

@overload
def atan2(
    y: FloatScalar, x: FloatScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.FloatScalar | lib.DoubleScalar: ...
@overload
def atan2(
    y: FloatArray, x: FloatArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.NumericArray[lib.FloatScalar] | lib.NumericArray[lib.DoubleScalar]: ...
@overload
def atan2(
    y: FloatScalar | FloatArray,
    x: FloatScalar | FloatArray,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.NumericArray[lib.FloatScalar] | lib.NumericArray[lib.DoubleScalar]: ...
@overload
def atan2(
    y: Expression | Any, x: Expression | Any, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...

# ========================= 2.5 Comparisons functions =========================
@overload
def equal(
    x: lib.Scalar, y: lib.Scalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.BooleanScalar: ...
@overload
def equal(
    x: lib.Scalar | lib.Array | lib.ChunkedArray,
    y: lib.Scalar | lib.Array | lib.ChunkedArray,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanArray: ...
@overload
def equal(
    x: Expression | Any,
    y: Expression | Any,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

greater = _clone_signature(equal)
greater_equal = _clone_signature(equal)
less = _clone_signature(equal)
less_equal = _clone_signature(equal)
not_equal = _clone_signature(equal)

@overload
def max_element_wise(
    *args: _ScalarT,
    skip_nulls: bool = True,
    options: ElementWiseAggregateOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _ScalarT: ...
@overload
def max_element_wise(
    *args: _ArrayT,
    skip_nulls: bool = True,
    options: ElementWiseAggregateOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _ArrayT: ...
@overload
def max_element_wise(
    *args: Expression,
    skip_nulls: bool = True,
    options: ElementWiseAggregateOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

min_element_wise = _clone_signature(equal)

# ========================= 2.6 Logical functions =========================
@overload
def and_(
    x: lib.BooleanScalar, y: lib.BooleanScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.BooleanScalar: ...
@overload
def and_(
    x: lib.BooleanScalar | BooleanArray,
    y: lib.BooleanScalar | BooleanArray,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanArray: ...
@overload
def and_(
    x: Expression | Any,
    y: Expression | Any,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

and_kleene = _clone_signature(and_)
and_not = _clone_signature(and_)
and_not_kleene = _clone_signature(and_)
or_ = _clone_signature(and_)
or_kleene = _clone_signature(and_)
xor = _clone_signature(and_)

@overload
def invert(
    x: lib.BooleanScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.BooleanScalar: ...
@overload
def invert(
    x: lib.BooleanArray,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanArray: ...
@overload
def invert(
    x: Expression,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

# ========================= 2.10 String predicates =========================
@overload
def ascii_is_alnum(
    strings: StringScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.BooleanScalar: ...
@overload
def ascii_is_alnum(
    strings: StringArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.BooleanArray: ...
@overload
def ascii_is_alnum(
    strings: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...

ascii_is_alpha = _clone_signature(ascii_is_alnum)
ascii_is_decimal = _clone_signature(ascii_is_alnum)
ascii_is_lower = _clone_signature(ascii_is_alnum)
ascii_is_printable = _clone_signature(ascii_is_alnum)
ascii_is_space = _clone_signature(ascii_is_alnum)
ascii_is_upper = _clone_signature(ascii_is_alnum)
utf8_is_alnum = _clone_signature(ascii_is_alnum)
utf8_is_alpha = _clone_signature(ascii_is_alnum)
utf8_is_decimal = _clone_signature(ascii_is_alnum)
utf8_is_digit = _clone_signature(ascii_is_alnum)
utf8_is_lower = _clone_signature(ascii_is_alnum)
utf8_is_numeric = _clone_signature(ascii_is_alnum)
utf8_is_printable = _clone_signature(ascii_is_alnum)
utf8_is_space = _clone_signature(ascii_is_alnum)
utf8_is_upper = _clone_signature(ascii_is_alnum)
ascii_is_title = _clone_signature(ascii_is_alnum)
utf8_is_title = _clone_signature(ascii_is_alnum)
string_is_ascii = _clone_signature(ascii_is_alnum)

# ========================= 2.11 String transforms =========================
@overload
def ascii_capitalize(
    strings: _StringScalarT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _StringScalarT: ...
@overload
def ascii_capitalize(
    strings: _StringArrayT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _StringArrayT: ...
@overload
def ascii_capitalize(
    strings: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...

ascii_lower = _clone_signature(ascii_capitalize)
ascii_reverse = _clone_signature(ascii_capitalize)
ascii_swapcase = _clone_signature(ascii_capitalize)
ascii_title = _clone_signature(ascii_capitalize)
ascii_upper = _clone_signature(ascii_capitalize)

@overload
def binary_length(
    strings: lib.BinaryScalar | lib.StringScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.Int32Scalar: ...
@overload
def binary_length(
    strings: lib.LargeBinaryScalar | lib.LargeStringScalar,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Scalar: ...
@overload
def binary_length(
    strings: lib.BinaryArray
    | lib.StringArray
    | lib.ChunkedArray[lib.BinaryScalar]
    | lib.ChunkedArray[lib.StringScalar],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int32Array: ...
@overload
def binary_length(
    strings: lib.LargeBinaryArray
    | lib.LargeStringArray
    | lib.ChunkedArray[lib.LargeBinaryScalar]
    | lib.ChunkedArray[lib.LargeStringScalar],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Array: ...
@overload
def binary_length(
    strings: Expression,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def binary_repeat(
    strings: _StringOrBinaryScalarT,
    num_repeats: int,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringOrBinaryScalarT: ...
@overload
def binary_repeat(
    strings: _StringOrBinaryScalarT,
    num_repeats: list[int] | list[int | None],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Array[_StringOrBinaryScalarT]: ...
@overload
def binary_repeat(
    strings: _StringOrBinaryArrayT,
    num_repeats: int | list[int] | list[int | None],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringOrBinaryArrayT: ...
@overload
def binary_repeat(
    strings: Expression,
    num_repeats: int | list[int] | list[int | None],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def binary_replace_slice(
    strings: _StringOrBinaryScalarT,
    /,
    start: int,
    stop: int,
    replacement: str | bytes,
    *,
    options: ReplaceSliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringOrBinaryScalarT: ...
@overload
def binary_replace_slice(
    strings: _StringOrBinaryArrayT,
    /,
    start: int,
    stop: int,
    replacement: str | bytes,
    *,
    options: ReplaceSliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringOrBinaryArrayT: ...
@overload
def binary_replace_slice(
    strings: Expression,
    /,
    start: int,
    stop: int,
    replacement: str | bytes,
    *,
    options: ReplaceSliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def binary_reverse(
    strings: _BinaryScalarT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _BinaryScalarT: ...
@overload
def binary_reverse(
    strings: _BinaryArrayT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _BinaryArrayT: ...
@overload
def binary_reverse(
    strings: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...
@overload
def replace_substring(
    strings: _StringScalarT,
    /,
    pattern: str | bytes,
    replacement: str | bytes,
    *,
    max_replacements: int | None = None,
    options: ReplaceSubstringOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringScalarT: ...
@overload
def replace_substring(
    strings: _StringArrayT,
    /,
    pattern: str | bytes,
    replacement: str | bytes,
    *,
    max_replacements: int | None = None,
    options: ReplaceSubstringOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringArrayT: ...
@overload
def replace_substring(
    strings: Expression,
    /,
    pattern: str | bytes,
    replacement: str | bytes,
    *,
    max_replacements: int | None = None,
    options: ReplaceSubstringOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

replace_substring_regex = _clone_signature(replace_substring)

@overload
def utf8_capitalize(
    strings: _StringScalarT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _StringScalarT: ...
@overload
def utf8_capitalize(
    strings: _StringArrayT, /, *, memory_pool: lib.MemoryPool | None = None
) -> _StringArrayT: ...
@overload
def utf8_capitalize(
    strings: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...
@overload
def utf8_length(
    strings: lib.StringScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.Int32Scalar: ...
@overload
def utf8_length(
    strings: lib.LargeStringScalar,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Scalar: ...
@overload
def utf8_length(
    strings: lib.StringArray | lib.ChunkedArray[lib.StringScalar],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int32Array: ...
@overload
def utf8_length(
    strings: lib.LargeStringArray | lib.ChunkedArray[lib.LargeStringScalar],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Array: ...
@overload
def utf8_length(
    strings: Expression,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

utf8_lower = _clone_signature(utf8_capitalize)

@overload
def utf8_replace_slice(
    strings: _StringScalarT,
    /,
    start: int,
    stop: int,
    replacement: str | bytes,
    *,
    options: ReplaceSliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringScalarT: ...
@overload
def utf8_replace_slice(
    strings: _StringArrayT,
    /,
    start: int,
    stop: int,
    replacement: str | bytes,
    *,
    options: ReplaceSliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringArrayT: ...
@overload
def utf8_replace_slice(
    strings: Expression,
    /,
    start: int,
    stop: int,
    replacement: str | bytes,
    *,
    options: ReplaceSliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

utf8_reverse = _clone_signature(utf8_capitalize)
utf8_swapcase = _clone_signature(utf8_capitalize)
utf8_title = _clone_signature(utf8_capitalize)
utf8_upper = _clone_signature(utf8_capitalize)

# ========================= 2.12 String padding =========================
@overload
def ascii_center(
    strings: _StringScalarT,
    /,
    width: int,
    padding: str = " ",
    lean_left_on_odd_padding: bool = True,
    *,
    options: PadOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringScalarT: ...
@overload
def ascii_center(
    strings: _StringArrayT,
    /,
    width: int,
    padding: str = " ",
    lean_left_on_odd_padding: bool = True,
    *,
    options: PadOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringArrayT: ...
@overload
def ascii_center(
    strings: Expression,
    /,
    width: int,
    padding: str = " ",
    lean_left_on_odd_padding: bool = True,
    *,
    options: PadOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

ascii_lpad = _clone_signature(ascii_center)
ascii_rpad = _clone_signature(ascii_center)
utf8_center = _clone_signature(ascii_center)
utf8_lpad = _clone_signature(ascii_center)
utf8_rpad = _clone_signature(ascii_center)

# ========================= 2.13 String trimming =========================
@overload
def ascii_ltrim(
    strings: _StringScalarT,
    /,
    characters: str,
    *,
    options: TrimOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringScalarT: ...
@overload
def ascii_ltrim(
    strings: _StringArrayT,
    /,
    characters: str,
    *,
    options: TrimOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringArrayT: ...
@overload
def ascii_ltrim(
    strings: Expression,
    /,
    characters: str,
    *,
    options: TrimOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

ascii_rtrim = _clone_signature(ascii_ltrim)
ascii_trim = _clone_signature(ascii_ltrim)
utf8_ltrim = _clone_signature(ascii_ltrim)
utf8_rtrim = _clone_signature(ascii_ltrim)
utf8_trim = _clone_signature(ascii_ltrim)

@overload
def ascii_ltrim_whitespace(
    strings: _StringScalarT,
    /,
    *,
    options: TrimOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringScalarT: ...
@overload
def ascii_ltrim_whitespace(
    strings: _StringArrayT,
    /,
    *,
    options: TrimOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringArrayT: ...
@overload
def ascii_ltrim_whitespace(
    strings: Expression,
    /,
    *,
    options: TrimOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

ascii_rtrim_whitespace = _clone_signature(ascii_ltrim_whitespace)
ascii_trim_whitespace = _clone_signature(ascii_ltrim_whitespace)
utf8_ltrim_whitespace = _clone_signature(ascii_ltrim_whitespace)
utf8_rtrim_whitespace = _clone_signature(ascii_ltrim_whitespace)
utf8_trim_whitespace = _clone_signature(ascii_ltrim_whitespace)

# ========================= 2.14 String splitting =========================
@overload
def ascii_split_whitespace(
    strings: _StringScalarT,
    /,
    *,
    max_splits: int | None = None,
    reverse: bool = False,
    options: SplitOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.ListArray[_StringScalarT]: ...
@overload
def ascii_split_whitespace(
    strings: lib.Array[lib.Scalar[_DataTypeT]],
    /,
    *,
    max_splits: int | None = None,
    reverse: bool = False,
    options: SplitOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.ListArray[lib.ListScalar[_DataTypeT]]: ...
@overload
def ascii_split_whitespace(
    strings: Expression,
    /,
    *,
    max_splits: int | None = None,
    reverse: bool = False,
    options: SplitOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def split_pattern(
    strings: _StringOrBinaryScalarT,
    /,
    pattern: str,
    *,
    max_splits: int | None = None,
    reverse: bool = False,
    options: SplitOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.ListArray[_StringOrBinaryScalarT]: ...
@overload
def split_pattern(
    strings: lib.Array[lib.Scalar[_DataTypeT]],
    /,
    pattern: str,
    *,
    max_splits: int | None = None,
    reverse: bool = False,
    options: SplitPatternOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.ListArray[lib.ListScalar[_DataTypeT]]: ...
@overload
def split_pattern(
    strings: Expression,
    /,
    pattern: str,
    *,
    max_splits: int | None = None,
    reverse: bool = False,
    options: SplitPatternOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

split_pattern_regex = _clone_signature(split_pattern)
utf8_split_whitespace = _clone_signature(ascii_split_whitespace)

# ========================= 2.15 String component extraction =========================
@overload
def extract_regex(
    strings: StringOrBinaryScalar,
    /,
    pattern: str,
    *,
    options: ExtractRegexOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.StructScalar: ...
@overload
def extract_regex(
    strings: StringOrBinaryArray,
    /,
    pattern: str,
    *,
    options: ExtractRegexOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.StructArray: ...
@overload
def extract_regex(
    strings: Expression,
    /,
    pattern: str,
    *,
    options: ExtractRegexOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

# ========================= 2.16 String join =========================
def binary_join(
    strings, separator, /, *, memory_pool: lib.MemoryPool | None = None
) -> StringScalar | StringArray: ...
@overload
def binary_join_element_wise(
    *strings: _StringOrBinaryScalarT,
    null_handling: Literal["emit_null", "skip", "replace"] = "emit_null",
    null_replacement: str = "",
    options: JoinOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringOrBinaryScalarT: ...
@overload
def binary_join_element_wise(
    *strings: _StringOrBinaryArrayT,
    null_handling: Literal["emit_null", "skip", "replace"] = "emit_null",
    null_replacement: str = "",
    options: JoinOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringOrBinaryArrayT: ...
@overload
def binary_join_element_wise(
    *strings: Expression,
    null_handling: Literal["emit_null", "skip", "replace"] = "emit_null",
    null_replacement: str = "",
    options: JoinOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

# ========================= 2.17 String Slicing =========================
@overload
def binary_slice(
    strings: _BinaryScalarT,
    /,
    start: int,
    stop: int | None = None,
    step: int = 1,
    *,
    options: SliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _BinaryScalarT: ...
@overload
def binary_slice(
    strings: _BinaryArrayT,
    /,
    start: int,
    stop: int | None = None,
    step: int = 1,
    *,
    options: SliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _BinaryArrayT: ...
@overload
def binary_slice(
    strings: Expression,
    /,
    start: int,
    stop: int | None = None,
    step: int = 1,
    *,
    options: SliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def utf8_slice_codeunits(
    strings: _StringScalarT,
    /,
    start: int,
    stop: int | None = None,
    step: int = 1,
    *,
    options: SliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringScalarT: ...
@overload
def utf8_slice_codeunits(
    strings: _StringArrayT,
    /,
    start: int,
    stop: int | None = None,
    step: int = 1,
    *,
    options: SliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _StringArrayT: ...
@overload
def utf8_slice_codeunits(
    strings: Expression,
    /,
    start: int,
    stop: int | None = None,
    step: int = 1,
    *,
    options: SliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

# ========================= 2.18 Containment tests =========================
@overload
def count_substring(
    strings: lib.StringScalar | lib.BinaryScalar,
    /,
    pattern: str,
    *,
    ignore_case: bool = False,
    options: MatchSubstringOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int32Scalar: ...
@overload
def count_substring(
    strings: lib.LargeStringScalar | lib.LargeBinaryScalar,
    /,
    pattern: str,
    *,
    ignore_case: bool = False,
    options: MatchSubstringOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Scalar: ...
@overload
def count_substring(
    strings: lib.StringArray
    | lib.BinaryArray
    | lib.ChunkedArray[lib.StringScalar]
    | lib.ChunkedArray[lib.BinaryScalar],
    /,
    pattern: str,
    *,
    ignore_case: bool = False,
    options: MatchSubstringOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int32Array: ...
@overload
def count_substring(
    strings: lib.LargeStringArray
    | lib.LargeBinaryArray
    | lib.ChunkedArray[lib.LargeStringScalar]
    | lib.ChunkedArray[lib.LargeBinaryScalar],
    /,
    pattern: str,
    *,
    ignore_case: bool = False,
    options: MatchSubstringOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Array: ...
@overload
def count_substring(
    strings: Expression,
    /,
    pattern: str,
    *,
    ignore_case: bool = False,
    options: MatchSubstringOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

count_substring_regex = _clone_signature(count_substring)

@overload
def ends_with(
    strings: StringScalar | BinaryScalar,
    /,
    pattern: str,
    *,
    ignore_case: bool = False,
    options: MatchSubstringOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanScalar: ...
@overload
def ends_with(
    strings: StringArray | BinaryArray,
    /,
    pattern: str,
    *,
    ignore_case: bool = False,
    options: MatchSubstringOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanArray: ...
@overload
def ends_with(
    strings: Expression,
    /,
    pattern: str,
    *,
    ignore_case: bool = False,
    options: MatchSubstringOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

find_substring = _clone_signature(count_substring)
find_substring_regex = _clone_signature(count_substring)

@overload
def index_in(
    values: lib.Scalar,
    /,
    value_set: lib.Array | lib.ChunkedArray,
    *,
    skip_nulls: bool = False,
    options: SetLookupOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int32Scalar: ...
@overload
def index_in(
    values: lib.Array | lib.ChunkedArray,
    /,
    value_set: lib.Array | lib.ChunkedArray,
    *,
    skip_nulls: bool = False,
    options: SetLookupOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int32Array: ...
@overload
def index_in(
    values: Expression,
    /,
    value_set: lib.Array | lib.ChunkedArray,
    *,
    skip_nulls: bool = False,
    options: SetLookupOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def is_in(
    values: lib.Scalar,
    /,
    value_set: lib.Array | lib.ChunkedArray,
    *,
    skip_nulls: bool = False,
    options: SetLookupOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanScalar: ...
@overload
def is_in(
    values: lib.Array | lib.ChunkedArray,
    /,
    value_set: lib.Array | lib.ChunkedArray,
    *,
    skip_nulls: bool = False,
    options: SetLookupOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanArray: ...
@overload
def is_in(
    values: Expression,
    /,
    value_set: lib.Array | lib.ChunkedArray,
    *,
    skip_nulls: bool = False,
    options: SetLookupOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

match_like = _clone_signature(ends_with)
match_substring = _clone_signature(ends_with)
match_substring_regex = _clone_signature(ends_with)
starts_with = _clone_signature(ends_with)

# ========================= 2.19 Categorizations =========================
@overload
def is_finite(
    values: NumericScalar | lib.NullScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.BooleanScalar: ...
@overload
def is_finite(
    values: NumericArray | lib.NullArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.BooleanArray: ...
@overload
def is_finite(
    values: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...

is_inf = _clone_signature(is_finite)
is_nan = _clone_signature(is_finite)

@overload
def is_null(
    values: lib.Scalar,
    /,
    *,
    nan_is_null: bool = False,
    options: NullOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanScalar: ...
@overload
def is_null(
    values: lib.Array | lib.ChunkedArray,
    /,
    *,
    nan_is_null: bool = False,
    options: NullOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanArray: ...
@overload
def is_null(
    values: Expression,
    /,
    *,
    nan_is_null: bool = False,
    options: NullOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def is_valid(
    values: lib.Scalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.BooleanScalar: ...
@overload
def is_valid(
    values: lib.Array | lib.ChunkedArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.BooleanArray: ...
@overload
def is_valid(
    values: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...

true_unless_null = _clone_signature(is_valid)

# ========================= 2.20 Selecting / multiplexing =========================
def case_when(cond, /, *cases, memory_pool: lib.MemoryPool | None = None): ...
def choose(indices, /, *values, memory_pool: lib.MemoryPool | None = None): ...
def coalesce(
    *values: _ScalarOrArrayT, memory_pool: lib.MemoryPool | None = None
) -> _ScalarOrArrayT: ...
def if_else(cond, left, right, /, *, memory_pool: lib.MemoryPool | None = None): ...

# ========================= 2.21 Structural transforms =========================

@overload
def list_value_length(
    lists: lib.ListArray | lib.ListViewArray | lib.FixedSizeListArray | lib.ChunkedArray,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int32Array: ...
@overload
def list_value_length(
    lists: lib.LargeListArray | lib.LargeListViewArray | lib.ChunkedArray,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Array: ...
@overload
def list_value_length(
    lists: Expression,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def make_struct(
    *args: lib.Scalar,
    field_names: list[str] | tuple[str, ...] = (),
    field_nullability: bool | None = None,
    field_metadata: list[lib.KeyValueMetadata] | None = None,
    options: MakeStructOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.StructScalar: ...
@overload
def make_struct(
    *args: lib.Array | lib.ChunkedArray,
    field_names: list[str] | tuple[str, ...] = (),
    field_nullability: bool | None = None,
    field_metadata: list[lib.KeyValueMetadata] | None = None,
    options: MakeStructOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.StructArray: ...
@overload
def make_struct(
    *args: Expression,
    field_names: list[str] | tuple[str, ...] = (),
    field_nullability: bool | None = None,
    field_metadata: list[lib.KeyValueMetadata] | None = None,
    options: MakeStructOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

# ========================= 2.22 Conversions =========================
@overload
def ceil_temporal(
    timestamps: _TemporalScalarT,
    /,
    multiple: int = 1,
    unit: Literal[
        "year",
        "quarter",
        "month",
        "week",
        "day",
        "hour",
        "minute",
        "second",
        "millisecond",
        "microsecond",
        "nanosecond",
    ] = "day",
    *,
    week_starts_monday: bool = True,
    ceil_is_strictly_greater: bool = False,
    calendar_based_origin: bool = False,
    options: RoundTemporalOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _TemporalScalarT: ...
@overload
def ceil_temporal(
    timestamps: _TemporalArrayT,
    /,
    multiple: int = 1,
    unit: Literal[
        "year",
        "quarter",
        "month",
        "week",
        "day",
        "hour",
        "minute",
        "second",
        "millisecond",
        "microsecond",
        "nanosecond",
    ] = "day",
    *,
    week_starts_monday: bool = True,
    ceil_is_strictly_greater: bool = False,
    calendar_based_origin: bool = False,
    options: RoundTemporalOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _TemporalArrayT: ...
@overload
def ceil_temporal(
    timestamps: Expression,
    /,
    multiple: int = 1,
    unit: Literal[
        "year",
        "quarter",
        "month",
        "week",
        "day",
        "hour",
        "minute",
        "second",
        "millisecond",
        "microsecond",
        "nanosecond",
    ] = "day",
    *,
    week_starts_monday: bool = True,
    ceil_is_strictly_greater: bool = False,
    calendar_based_origin: bool = False,
    options: RoundTemporalOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

floor_temporal = _clone_signature(ceil_temporal)
round_temporal = _clone_signature(ceil_temporal)

@overload
def cast(
    arr: lib.Scalar,
    target_type: _DataTypeT,
    safe: bool | None = None,
    options: CastOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Scalar[_DataTypeT]: ...
@overload
def cast(
    arr: lib.Array,
    target_type: _DataTypeT,
    safe: bool | None = None,
    options: CastOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Array[lib.Scalar[_DataTypeT]]: ...
@overload
def cast(
    arr: lib.ChunkedArray,
    target_type: _DataTypeT,
    safe: bool | None = None,
    options: CastOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.ChunkedArray[lib.Scalar[_DataTypeT]]: ...
@overload
def strftime(
    timestamps: TemporalScalar,
    /,
    format: str = "%Y-%m-%dT%H:%M:%S",
    locale: str = "C",
    *,
    options: StrftimeOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.StringScalar: ...
@overload
def strftime(
    timestamps: TemporalArray,
    /,
    format: str = "%Y-%m-%dT%H:%M:%S",
    locale: str = "C",
    *,
    options: StrftimeOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.StringArray: ...
@overload
def strftime(
    timestamps: Expression,
    /,
    format: str = "%Y-%m-%dT%H:%M:%S",
    locale: str = "C",
    *,
    options: StrftimeOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def strptime(
    strings: StringScalar,
    /,
    format: str,
    unit: Literal["s", "ms", "us", "ns"],
    error_is_null: bool = False,
    *,
    options: StrptimeOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.TimestampScalar: ...
@overload
def strptime(
    strings: StringArray,
    /,
    format: str,
    unit: Literal["s", "ms", "us", "ns"],
    error_is_null: bool = False,
    *,
    options: StrptimeOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.TimestampArray: ...
@overload
def strptime(
    strings: Expression,
    /,
    format: str,
    unit: Literal["s", "ms", "us", "ns"],
    error_is_null: bool = False,
    *,
    options: StrptimeOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

# ========================= 2.23 Temporal component extraction =========================
@overload
def day(
    values: TemporalScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.Int64Scalar: ...
@overload
def day(
    values: TemporalArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.Int64Array: ...
@overload
def day(values: Expression, /, *, memory_pool: lib.MemoryPool | None = None) -> Expression: ...
@overload
def day_of_week(
    values: TemporalScalar,
    /,
    *,
    count_from_zero: bool = True,
    week_start: int = 1,
    options: DayOfWeekOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Scalar: ...
@overload
def day_of_week(
    values: TemporalArray,
    /,
    *,
    count_from_zero: bool = True,
    week_start: int = 1,
    options: DayOfWeekOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Array: ...
@overload
def day_of_week(
    values: Expression,
    /,
    *,
    count_from_zero: bool = True,
    week_start: int = 1,
    options: DayOfWeekOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

day_of_year = _clone_signature(day)

@overload
def hour(
    values: lib.TimestampScalar | lib.Time32Scalar | lib.Time64Scalar,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Scalar: ...
@overload
def hour(
    values: lib.TimestampArray
    | lib.Time32Array
    | lib.Time64Array
    | lib.ChunkedArray[lib.TimestampScalar]
    | lib.ChunkedArray[lib.Time32Scalar]
    | lib.ChunkedArray[lib.Time64Scalar],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Array: ...
@overload
def hour(
    values: Expression,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def is_dst(
    values: lib.TimestampScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.BooleanScalar: ...
@overload
def is_dst(
    values: lib.TimestampArray | lib.ChunkedArray[lib.TimestampScalar],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanArray: ...
@overload
def is_dst(values: Expression, /, *, memory_pool: lib.MemoryPool | None = None) -> Expression: ...
@overload
def iso_week(
    values: lib.TimestampScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.Int64Scalar: ...
@overload
def iso_week(
    values: lib.TimestampArray | lib.ChunkedArray[lib.TimestampScalar],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Array: ...
@overload
def iso_week(
    values: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...

iso_year = _clone_signature(iso_week)

@overload
def is_leap_year(
    values: lib.TimestampScalar | lib.Date32Scalar | lib.Date64Scalar,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanScalar: ...
@overload
def is_leap_year(
    values: lib.TimestampArray
    | lib.Date32Array
    | lib.Date64Array
    | lib.ChunkedArray[lib.TimestampScalar]
    | lib.ChunkedArray[lib.Date32Scalar]
    | lib.ChunkedArray[lib.Date64Scalar],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.BooleanArray: ...
@overload
def is_leap_year(
    values: Expression,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

microsecond = _clone_signature(iso_week)
millisecond = _clone_signature(iso_week)
minute = _clone_signature(iso_week)
month = _clone_signature(day_of_week)
nanosecond = _clone_signature(hour)
quarter = _clone_signature(day_of_week)
second = _clone_signature(hour)
subsecond = _clone_signature(hour)
us_week = _clone_signature(iso_week)
us_year = _clone_signature(iso_week)
year = _clone_signature(iso_week)

@overload
def week(
    values: lib.TimestampScalar,
    /,
    *,
    week_starts_monday: bool = True,
    count_from_zero: bool = False,
    first_week_is_fully_in_year: bool = False,
    options: WeekOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Scalar: ...
@overload
def week(
    values: lib.TimestampArray | lib.ChunkedArray[lib.TimestampScalar],
    /,
    *,
    week_starts_monday: bool = True,
    count_from_zero: bool = False,
    first_week_is_fully_in_year: bool = False,
    options: WeekOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Array: ...
@overload
def week(
    values: Expression,
    /,
    *,
    week_starts_monday: bool = True,
    count_from_zero: bool = False,
    first_week_is_fully_in_year: bool = False,
    options: WeekOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def year_month_day(
    values: TemporalScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.StructScalar: ...
@overload
def year_month_day(
    values: TemporalArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.StructArray: ...
@overload
def year_month_day(
    values: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...

# ========================= 2.24 Temporal difference =========================
def day_time_interval_between(start, end, /, *, memory_pool: lib.MemoryPool | None = None): ...
def days_between(
    start, end, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.Int64Scalar | lib.Int64Array: ...

hours_between = _clone_signature(days_between)
microseconds_between = _clone_signature(days_between)
milliseconds_between = _clone_signature(days_between)
minutes_between = _clone_signature(days_between)

def month_day_nano_interval_between(
    start, end, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.MonthDayNanoIntervalScalar | lib.MonthDayNanoIntervalArray: ...
def month_interval_between(start, end, /, *, memory_pool: lib.MemoryPool | None = None): ...

nanoseconds_between = _clone_signature(days_between)
quarters_between = _clone_signature(days_between)
seconds_between = _clone_signature(days_between)

def weeks_between(
    start,
    end,
    /,
    *,
    count_from_zero: bool = True,
    week_start: int = 1,
    options: DayOfWeekOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Int64Scalar | lib.Int64Array: ...

years_between = _clone_signature(days_between)

# ========================= 2.25 Timezone handling =========================
@overload
def assume_timezone(
    timestamps: lib.TimestampScalar,
    /,
    timezone: str,
    *,
    ambiguous: Literal["raise", "earliest", "latest"] = "raise",
    nonexistent: Literal["raise", "earliest", "latest"] = "raise",
    options: AssumeTimezoneOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.TimestampScalar: ...
@overload
def assume_timezone(
    timestamps: lib.TimestampArray | lib.ChunkedArray[lib.TimestampScalar],
    /,
    timezone: str,
    *,
    ambiguous: Literal["raise", "earliest", "latest"] = "raise",
    nonexistent: Literal["raise", "earliest", "latest"] = "raise",
    options: AssumeTimezoneOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.TimestampArray: ...
@overload
def assume_timezone(
    timestamps: Expression,
    /,
    timezone: str,
    *,
    ambiguous: Literal["raise", "earliest", "latest"] = "raise",
    nonexistent: Literal["raise", "earliest", "latest"] = "raise",
    options: AssumeTimezoneOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def local_timestamp(
    timestamps: lib.TimestampScalar, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.TimestampScalar: ...
@overload
def local_timestamp(
    timestamps: lib.TimestampArray | lib.ChunkedArray[lib.TimestampScalar],
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.TimestampArray: ...
@overload
def local_timestamp(
    timestamps: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...

# ========================= 2.26 Random number generation =========================
def random(
    n: int,
    *,
    initializer: Literal["system"] | int = "system",
    options: RandomOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.DoubleArray: ...

# ========================= 3. Array-wise (“vector”) functions =========================

# ========================= 3.1 Cumulative Functions =========================
@overload
def cumulative_sum(
    values: _NumericArrayT,
    /,
    start: lib.Scalar | None = None,
    *,
    skip_nulls: bool = False,
    options: CumulativeSumOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericArrayT: ...
@overload
def cumulative_sum(
    values: Expression,
    /,
    start: lib.Scalar | None = None,
    *,
    skip_nulls: bool = False,
    options: CumulativeSumOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

cumulative_sum_checked = _clone_signature(cumulative_sum)
cumulative_prod = _clone_signature(cumulative_sum)
cumulative_prod_checked = _clone_signature(cumulative_sum)
cumulative_max = _clone_signature(cumulative_sum)
cumulative_min = _clone_signature(cumulative_sum)
cumulative_mean = _clone_signature(cumulative_sum)

# ========================= 3.2 Associative transforms =========================

@overload
def dictionary_encode(
    array: _ScalarOrArrayT,
    /,
    null_encoding: Literal["mask", "encode"] = "mask",
    *,
    options=None,
    memory_pool: lib.MemoryPool | None = None,
) -> _ScalarOrArrayT: ...
@overload
def dictionary_encode(
    array: Expression,
    /,
    null_encoding: Literal["mask", "encode"] = "mask",
    *,
    options=None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def unique(array: _ArrayT, /, *, memory_pool: lib.MemoryPool | None = None) -> _ArrayT: ...
@overload
def unique(array: Expression, /, *, memory_pool: lib.MemoryPool | None = None) -> Expression: ...
@overload
def value_counts(
    array: lib.Array | lib.ChunkedArray, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.StructArray: ...
@overload
def value_counts(
    array: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...

# ========================= 3.3 Selections =========================
@overload
def array_filter(
    array: _ArrayT,
    selection_filter: list[bool] | list[bool | None] | BooleanArray,
    /,
    null_selection_behavior: Literal["drop", "emit_null"] = "drop",
    *,
    options: FilterOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _ArrayT: ...
@overload
def array_filter(
    array: Expression,
    selection_filter: list[bool] | list[bool | None] | BooleanArray,
    /,
    null_selection_behavior: Literal["drop", "emit_null"] = "drop",
    *,
    options: FilterOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def array_take(
    array: _ArrayT,
    indices: list[int]
    | list[int | None]
    | lib.Int16Array
    | lib.Int32Array
    | lib.Int64Array
    | lib.ChunkedArray[lib.Int16Scalar]
    | lib.ChunkedArray[lib.Int32Scalar]
    | lib.ChunkedArray[lib.Int64Scalar],
    /,
    *,
    boundscheck: bool = True,
    options: TakeOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _ArrayT: ...
@overload
def array_take(
    array: Expression,
    indices: list[int]
    | list[int | None]
    | lib.Int16Array
    | lib.Int32Array
    | lib.Int64Array
    | lib.ChunkedArray[lib.Int16Scalar]
    | lib.ChunkedArray[lib.Int32Scalar]
    | lib.ChunkedArray[lib.Int64Scalar],
    /,
    *,
    boundscheck: bool = True,
    options: TakeOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def drop_null(input: _ArrayT, /, *, memory_pool: lib.MemoryPool | None = None) -> _ArrayT: ...
@overload
def drop_null(
    input: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...

filter = array_filter
take = array_take

# ========================= 3.4 Containment tests  =========================
@overload
def indices_nonzero(
    values: lib.BooleanArray
    | lib.NullArray
    | NumericArray
    | lib.Decimal128Array
    | lib.Decimal256Array,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.UInt64Array: ...
@overload
def indices_nonzero(
    values: Expression,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

# ========================= 3.5 Sorts and partitions  =========================
@overload
def array_sort_indices(
    array: lib.Array | lib.ChunkedArray,
    /,
    order: Literal["ascending", "descending"] = "ascending",
    *,
    null_placement: Literal["at_start", "at_end"] = "at_end",
    options: ArraySortOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.UInt64Array: ...
@overload
def array_sort_indices(
    array: Expression,
    /,
    order: Literal["ascending", "descending"] = "ascending",
    *,
    null_placement: Literal["at_start", "at_end"] = "at_end",
    options: ArraySortOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def partition_nth_indices(
    array: lib.Array | lib.ChunkedArray,
    /,
    pivot: int,
    *,
    null_placement: Literal["at_start", "at_end"] = "at_end",
    options: PartitionNthOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.UInt64Array: ...
@overload
def partition_nth_indices(
    array: Expression,
    /,
    pivot: int,
    *,
    null_placement: Literal["at_start", "at_end"] = "at_end",
    options: PartitionNthOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
def rank(
    input: lib.Array | lib.ChunkedArray,
    /,
    sort_keys: Literal["ascending", "descending"] = "ascending",
    *,
    null_placement: Literal["at_start", "at_end"] = "at_end",
    tiebreaker: Literal["min", "max", "first", "dense"] = "first",
    options: RankOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.UInt64Array: ...
@overload
def select_k_unstable(
    input: lib.Array | lib.ChunkedArray,
    /,
    k: int,
    sort_keys: list[tuple[str, Literal["ascending", "descending"]]],
    *,
    options: SelectKOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.UInt64Array: ...
@overload
def select_k_unstable(
    input: Expression,
    /,
    k: int,
    sort_keys: list[tuple[str, Literal["ascending", "descending"]]],
    *,
    options: SelectKOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def sort_indices(
    input: lib.Array | lib.ChunkedArray | lib.RecordBatch | lib.Table,
    /,
    sort_keys: Sequence[tuple[str, Literal["ascending", "descending"]]],
    *,
    null_placement: Literal["at_start", "at_end"] = "at_end",
    options: SortOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.UInt64Array: ...
@overload
def sort_indices(
    input: Expression,
    /,
    sort_keys: Sequence[tuple[str, Literal["ascending", "descending"]]],
    *,
    null_placement: Literal["at_start", "at_end"] = "at_end",
    options: SortOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

# ========================= 3.6 Structural transforms =========================
@overload
def list_element(
    lists: Expression, index, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...
@overload
def list_element(
    lists, index, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.ListArray: ...
@overload
def list_flatten(
    lists: Expression,
    /,
    recursive: bool = False,
    *,
    options: ListFlattenOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def list_flatten(
    lists,
    /,
    recursive: bool = False,
    *,
    options: ListFlattenOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.ListArray: ...
@overload
def list_parent_indices(
    lists: Expression, /, *, memory_pool: lib.MemoryPool | None = None
) -> Expression: ...
@overload
def list_parent_indices(
    lists, /, *, memory_pool: lib.MemoryPool | None = None
) -> lib.Int64Array: ...
@overload
def list_slice(
    lists: Expression,
    /,
    start: int,
    stop: int | None = None,
    step: int = 1,
    return_fixed_size_list: bool | None = None,
    *,
    options: ListSliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...
@overload
def list_slice(
    lists,
    /,
    start: int,
    stop: int | None = None,
    step: int = 1,
    return_fixed_size_list: bool | None = None,
    *,
    options: ListSliceOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.ListArray: ...
def map_lookup(
    container,
    /,
    query_key,
    occurrence: str,
    *,
    options: MapLookupOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
): ...
def struct_field(
    values,
    /,
    indices,
    *,
    options: StructFieldOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
): ...
def fill_null_backward(values, /, *, memory_pool: lib.MemoryPool | None = None): ...
def fill_null_forward(values, /, *, memory_pool: lib.MemoryPool | None = None): ...
def replace_with_mask(
    values,
    mask: list[bool] | list[bool | None] | BooleanArray,
    replacements,
    /,
    *,
    memory_pool: lib.MemoryPool | None = None,
): ...

# ========================= 3.7 Pairwise functions =========================
@overload
def pairwise_diff(
    input: _NumericOrTemporalArrayT,
    /,
    period: int = 1,
    *,
    options: PairwiseOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> _NumericOrTemporalArrayT: ...
@overload
def pairwise_diff(
    input: Expression,
    /,
    period: int = 1,
    *,
    options: PairwiseOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> Expression: ...

pairwise_diff_checked = _clone_signature(pairwise_diff)



from dataclasses import dataclass, field
from typing import IO, Callable, Literal

from _typeshed import StrPath

from . import lib

@dataclass(kw_only=True)
class ReadOptions(lib._Weakrefable):
    use_threads: bool = field(default=True, kw_only=False)
    block_size: int | None = None
    skip_rows: int = 0
    skip_rows_after_names: int = 0
    column_names: list[str] | None = None
    autogenerate_column_names: bool = False
    encoding: str = "utf8"

    def validate(self) -> None: ...

@dataclass(kw_only=True)
class ParseOptions(lib._Weakrefable):
    delimiter: str = field(default=",", kw_only=False)
    quote_char: str | Literal[False] = '"'
    double_quote: bool = True
    escape_char: str | Literal[False] = False
    newlines_in_values: bool = False
    ignore_empty_lines: bool = True
    invalid_row_handler: Callable[[InvalidRow], Literal["skip", "error"]] | None = None

    def validate(self) -> None: ...

@dataclass(kw_only=True)
class ConvertOptions(lib._Weakrefable):
    check_utf8: bool = field(default=True, kw_only=False)
    column_types: lib.Schema | dict | None = None
    null_values: list[str] | None = None
    true_values: list[str] | None = None
    false_values: list[str] | None = None
    decimal_point: str = "."
    strings_can_be_null: bool = False
    quoted_strings_can_be_null: bool = True
    include_columns: list[str] | None = None
    include_missing_columns: bool = False
    auto_dict_encode: bool = False
    auto_dict_max_cardinality: int | None = None
    timestamp_parsers: list[str] | None = None

    def validate(self) -> None: ...

@dataclass(kw_only=True)
class WriteOptions(lib._Weakrefable):
    include_header: bool = field(default=True, kw_only=False)
    batch_size: int = 1024
    delimiter: str = ","
    quoting_style: Literal["needed", "all_valid", "none"] = "needed"

    def validate(self) -> None: ...

@dataclass
class InvalidRow(lib._Weakrefable):
    expected_columns: int
    actual_columns: int
    number: int | None
    text: str

class CSVWriter(lib._CRecordBatchWriter):
    def __init__(
        self,
        # TODO: OutputStream
        sink: StrPath | IO,
        schema: lib.Schema,
        write_options: WriteOptions | None = None,
        *,
        memory_pool: lib.MemoryPool | None = None,
    ) -> None: ...

class CSVStreamingReader(lib.RecordBatchReader): ...

ISO8601: lib._Weakrefable

def open_csv(
    input_file: StrPath | IO,
    read_options: ReadOptions | None = None,
    parse_options: ParseOptions | None = None,
    convert_options: ConvertOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> CSVStreamingReader: ...
def read_csv(
    input_file: StrPath | IO,
    read_options: ReadOptions | None = None,
    parse_options: ParseOptions | None = None,
    convert_options: ConvertOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> lib.Table: ...
def write_csv(
    data: lib.RecordBatch | lib.Table,
    output_file: StrPath | lib.NativeFile | IO,
    write_options: WriteOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> None: ...



import datetime as dt

from ._fs import FileSystem
from .lib import KeyValueMetadata

class GcsFileSystem(FileSystem):
    def __init__(
        self,
        *,
        anonymous: bool = False,
        access_token: str | None = None,
        target_service_account: str | None = None,
        credential_token_expiration: dt.datetime | None = None,
        default_bucket_location: str = "US",
        scheme: str = "https",
        endpoint_override: str | None = None,
        default_metadata: dict | KeyValueMetadata | None = None,
        retry_time_limit: dt.timedelta | None = None,
        project_id: str | None = None,
    ): ...
    @property
    def default_bucket_location(self) -> str: ...
    @property
    def project_id(self) -> str: ...
